"""
 File:  
 midifile.py
 
 Contents and purpose:
 Utilities used throughout JetCreator
 
 Copyright (c) 2008 Android Open Source Project
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

import logging
import struct
import copy
import array

# JET events
JET_EVENT_MARKER = 102
JET_MARKER_LOOP_END = 0
JET_EVENT_TRIGGER_CLIP = 103

# header definitions
SMF_HEADER_FMT = '>4slHHH'
SMF_RIFF_TAG = 'MThd'

SMF_TRACK_HEADER_FMT = '>4sl'
SMF_TRACK_RIFF_TAG = 'MTrk'

# defaults
DEFAULT_PPQN = 120
DEFAULT_BEATS_PER_MEASURE = 4
DEFAULT_TIME_FORMAT = '%03d:%02d:%03d'

# force note-offs to end of list
MAX_SEQ_NUM = 0x7fffffff

# MIDI messages
NOTE_OFF = 0x80
NOTE_ON = 0x90
POLY_KEY_PRESSURE = 0xa0
CONTROL_CHANGE = 0xb0
PROGRAM_CHANGE = 0xc0
CHANNEL_PRESSURE = 0xd0
PITCH_BEND = 0xe0

# System common messages
SYSEX = 0xf0
MIDI_TIME_CODE = 0xf1
SONG_POSITION_POINTER = 0xf2
SONG_SELECT = 0xf3
RESERVED_F4 = 0xf4
RESERVED_F5 = 0xf5
TUNE_REQUEST = 0xf6
END_SYSEX = 0xf7

# System real-time messages
TIMING_CLOCK = 0xf8
RESERVED_F9 = 0xf9
START = 0xfa
CONTINUE = 0xfb
STOP = 0xfc
RESERVED_FD = 0xfd
ACTIVE_SENSING = 0xfe
SYSTEM_RESET = 0xff

ONE_BYTE_MESSAGES = (
	TUNE_REQUEST,
	TIMING_CLOCK, 
	RESERVED_F9, 
	START, 
	CONTINUE, 
	STOP,
	RESERVED_FD,
	ACTIVE_SENSING,
	SYSTEM_RESET)

THREE_BYTE_MESSAGES = (
	NOTE_OFF, 
	NOTE_ON, 
	POLY_KEY_PRESSURE, 
	CONTROL_CHANGE, 
	PITCH_BEND)

MIDI_MESSAGES = (
	NOTE_OFF, 
	NOTE_ON, 
	POLY_KEY_PRESSURE, 
	CONTROL_CHANGE, 
	CHANNEL_PRESSURE, 
	PITCH_BEND, 
	SYSEX)

# Meta-events
META_EVENT = 0xff
META_EVENT_SEQUENCE_NUMBER = 0x00
META_EVENT_TEXT_EVENT = 0x01
META_EVENT_COPYRIGHT_NOTICE = 0x02
META_EVENT_SEQUENCE_TRACK_NAME = 0x03
META_EVENT_INSTRUMENT_NAME = 0x04
META_EVENT_LYRIC = 0x05
META_EVENT_MARKER = 0x06
META_EVENT_CUE_POINT = 0x07
META_EVENT_MIDI_CHANNEL_PREFIX = 0x20
META_EVENT_END_OF_TRACK = 0x2f
META_EVENT_SET_TEMPO = 0x51
META_EVENT_SMPTE_OFFSET = 0x54
META_EVENT_TIME_SIGNATURE = 0x58
META_EVENT_KEY_SIGNATURE = 0x59
META_EVENT_SEQUENCER_SPECIFIC = 0x7f

# recurring error messages
MSG_NOT_SMF_FILE = 'Not an SMF file - aborting parse!'
MSG_INVALID_TRACK_HEADER = 'Track header is invalid'
MSG_TYPE_MISMATCH = 'msg_type does not match event type'

LARGE_TICK_WARNING = 1000

# default control values
CTRL_BANK_SELECT_MSB = 0
CTRL_MOD_WHEEL = 1
CTRL_RPN_DATA_MSB = 6
CTRL_VOLUME = 7
CTRL_PAN = 10
CTRL_EXPRESSION = 11
CTRL_BANK_SELECT_LSB = 32
CTRL_RPN_DATA_LSB = 38
CTRL_SUSTAIN = 64
CTRL_RPN_LSB = 100
CTRL_RPN_MSB = 101
CTRL_RESET_CONTROLLERS = 121

RPN_PITCH_BEND_SENSITIVITY = 0
RPN_FINE_TUNING = 1
RPN_COARSE_TUNING = 2

MONITOR_CONTROLLERS = (
	CTRL_BANK_SELECT_MSB, 
	CTRL_MOD_WHEEL, 
	CTRL_RPN_DATA_MSB,
	CTRL_VOLUME, 
	CTRL_PAN, 
	CTRL_EXPRESSION, 
	CTRL_BANK_SELECT_LSB,
	CTRL_RPN_DATA_LSB,
	CTRL_SUSTAIN,
	CTRL_RPN_LSB,
	CTRL_RPN_MSB)

MONITOR_RPNS = (
	RPN_PITCH_BEND_SENSITIVITY,
	RPN_FINE_TUNING,
	RPN_COARSE_TUNING)

RPN_PITCH_BEND_SENSITIVITY = 0
RPN_FINE_TUNING = 1
RPN_COARSE_TUNING = 2

DEFAULT_CONTROLLER_VALUES = {
	CTRL_BANK_SELECT_MSB : 121,
	CTRL_MOD_WHEEL : 0,
	CTRL_RPN_DATA_MSB : 0,
	CTRL_VOLUME : 100,
	CTRL_PAN : 64,
	CTRL_EXPRESSION : 127,
	CTRL_RPN_DATA_LSB : 0,
	CTRL_BANK_SELECT_LSB : 0,
	CTRL_SUSTAIN : 0,
	CTRL_RPN_LSB : 0x7f,
	CTRL_RPN_MSB : 0x7f}

DEFAULT_RPN_VALUES = {
	RPN_PITCH_BEND_SENSITIVITY : 0x100,
	RPN_FINE_TUNING : 0,
	RPN_COARSE_TUNING : 1}

# initialize logger
midi_file_logger = logging.getLogger('MIDI_file')
midi_file_logger.setLevel(logging.NOTSET)


class trackGrid(object):
	def __init__ (self, track, channel, name, empty):
		self.track = track
		self.channel = channel
		self.name = name
		self.empty = empty
	def __str__ (self):
		return "['%s', '%s', '%s']" % (self.track, self.channel, self.name)
		

#---------------------------------------------------------------
# MIDIFileException
#---------------------------------------------------------------
class MIDIFileException (Exception):
	def __init__ (self, stream, msg):
		stream.error_loc = stream.tell()
		self.stream = stream
		self.msg = msg
	def __str__ (self):
		return '[%d]: %s' % (self.stream.error_loc, self.msg)

#---------------------------------------------------------------
# TimeBase
#---------------------------------------------------------------
class TimeBase (object):
	def __init__ (self, ppqn=DEFAULT_PPQN, beats_per_measure=DEFAULT_BEATS_PER_MEASURE):
		self.ppqn = ppqn
		self.beats_per_measure = beats_per_measure

	def ConvertToTicks (self, measures, beats, ticks):
		total_beats = beats + (measures * self.beats_per_measure)
		total_ticks = ticks + (total_beats * self.ppqn)
		return total_ticks

	def ConvertTicksToMBT (self, ticks):
		beats = ticks / self.ppqn
		ticks -= beats * self.ppqn
		measures = beats / self.beats_per_measure
		beats -= measures * self.beats_per_measure
		return (measures, beats, ticks)

	def ConvertTicksToStr (self, ticks, format=DEFAULT_TIME_FORMAT):
		measures, beats, ticks = self.ConvertTicksToMBT(ticks)
		return format % (measures, beats, ticks)

	def ConvertStrTimeToTuple(self, s):
		try:
			measures, beats, ticks = s.split(':',3)
			return (int(measures), int(beats), int(ticks))
		except:
			return (0,0,0)

	def ConvertStrTimeToTicks(self, s):
		measures, beats, ticks = self.ConvertStrTimeToTuple(s)
		return self.ConvertToTicks(measures, beats, ticks)
	
	def MbtDifference(self, mbt1, mbt2):
		t1 = self.ConvertToTicks(mbt1[0], mbt1[1], mbt1[2])
		t2 = self.ConvertToTicks(mbt2[0], mbt2[1], mbt2[2])
		return abs(t1-t2)
	
		
#---------------------------------------------------------------
# Helper functions
#---------------------------------------------------------------
def ReadByte (stream):
	try:
		return ord(stream.read(1))
	except TypeError:
		stream.error_loc = stream.tell()
		raise MIDIFileException(stream, 'Unexpected EOF')

def ReadBytes (stream, length):
	bytes = []
	for i in range(length):
		bytes.append(ReadByte(stream))
	return bytes

def ReadVarLenQty (stream):
	value = 0
	while 1:
		byte = ReadByte(stream)
		value = (value << 7) + (byte & 0x7f)
		if byte & 0x80 == 0:
			return value

def WriteByte (stream, value):
	stream.write(chr(value))
	
def WriteBytes (stream, bytes):
	for byte in bytes:
		WriteByte(stream, byte)			
	
def WriteVarLenQty (stream, value):
	bytes = [value & 0x7f]
	value = value >> 7
	while value > 0:
		bytes.append((value & 0x7f) | 0x80)
		value = value >> 7
	bytes.reverse()
	WriteBytes(stream, bytes)

#---------------------------------------------------------------
# EventFilter
#---------------------------------------------------------------
class EventFilter (object):
	pass

class EventTypeFilter (object):
	def __init__ (self, events, exclude=True):
		self.events = events
		self.exclude = exclude
	def Check (self, event):
		if event.msg_type in self.events:
			return not self.exclude
		return self.exclude
	
class NoteFilter (EventFilter):
	def __init__ (self, notes, exclude=True):
		self.notes = notes
		self.exclude = exclude
	def Check (self, event):
		if event.msg_type in (NOTE_ON, NOTE_OFF):
			if event.note in self.notes:
				return not self.exclude
		return self.exclude

class ChannelFilter (EventFilter):
	def __init__ (self, channel, exclude=True):
		self.channel = channel
		self.exclude = exclude
	def Check (self, event):
		if event.msg_type in (NOTE_ON, NOTE_OFF, POLY_KEY_PRESSURE, CONTROL_CHANGE, CHANNEL_PRESSURE, PITCH_BEND):
			if event.channel in self.channel:
				return not self.exclude
		return self.exclude

#---------------------------------------------------------------
# MIDIEvent
#---------------------------------------------------------------
class MIDIEvent (object):
	"""Factory for creating MIDI events from a stream."""
	@staticmethod
	def ReadFromStream (stream, seq, ticks, msg_type):
		if msg_type == SYSEX:
			return SysExEvent.ReadFromStream(stream, seq, ticks, msg_type)
		elif msg_type == END_SYSEX:
			return SysExContEvent.ReadFromStream(stream, seq, ticks, msg_type)
		elif msg_type == META_EVENT:
			return MetaEvent.ReadFromStream(stream, seq, ticks, msg_type)
		else:
			high_nibble = msg_type & 0xf0
			if high_nibble == NOTE_OFF:
				return NoteOffEvent.ReadFromStream(stream, seq, ticks, msg_type)
			elif high_nibble == NOTE_ON:
				return NoteOnEvent.ReadFromStream(stream, seq, ticks, msg_type)
			elif high_nibble == POLY_KEY_PRESSURE:
				return PolyKeyPressureEvent.ReadFromStream(stream, seq, ticks, msg_type)
			elif high_nibble == CONTROL_CHANGE:
				return ControlChangeEvent.ReadFromStream(stream, seq, ticks, msg_type)
			elif high_nibble == PROGRAM_CHANGE:
				return ProgramChangeEvent.ReadFromStream(stream, seq, ticks, msg_type)
			elif high_nibble == CHANNEL_PRESSURE:
				return ChannelPressureEvent.ReadFromStream(stream, seq, ticks, msg_type)
			elif high_nibble == PITCH_BEND:
				return PitchBendEvent.ReadFromStream(stream, seq, ticks, msg_type)
			else:
				stream.Warning('Ignoring unexpected message type 0x%02x' % msg_type)
	def WriteTicks (self, stream, track):
		WriteVarLenQty(stream, self.ticks - track.ticks)
		track.ticks = self.ticks
	def WriteRunningStatus (self, stream, track, filters, msg, data1, data2=None):
		if not self.CheckFilters(filters):
			return
		self.WriteTicks(stream, track)
		status = msg + self.channel
		if track.running_status != status:
			WriteByte(stream, status)
			track.running_status = status
		WriteByte(stream, data1)
		if data2 is not None:
			WriteByte(stream, data2)
	def CheckFilters (self, filters):
		if filters is None or not len(filters):
			return True
			
		# never filter meta-events
		if (self.msg_type == META_EVENT) and (self.meta_type == META_EVENT_END_OF_TRACK):
			return True

		# check all filters 
		for f in filters:
			if not f.Check(self):
				return False
		return True

	def TimeEventStr (self, timebase):
		return '[%s]: %s' % (timebase.ConvertTicksToStr(self.ticks), self.__str__())

#---------------------------------------------------------------
# NoteOffEvent
#---------------------------------------------------------------
class NoteOffEvent (MIDIEvent):
	def __init__ (self, ticks, seq, channel, note, velocity):
		self.name = 'NoteOff'
		self.msg_type = NOTE_OFF
		self.seq = seq
		self.ticks = ticks
		self.channel = channel
		self.note = note
		self.velocity = velocity
	@staticmethod
	def ReadFromStream (stream, seq, ticks, msg_type):
		ticks = ticks
		channel = msg_type & 0x0f
		note = ReadByte(stream)
		velocity = ReadByte(stream)
		if msg_type & 0xf0 != NOTE_OFF:
			stream.seek(-2,1)
			raise MIDIFileException(stream, MSG_TYPE_MISMATCH)
		return NoteOffEvent(ticks, seq, channel, note, velocity)
	def WriteToStream (self, stream, track, filters=None):
		# special case for note-off using zero velocity
		if self.velocity > 0:
			self.WriteRunningStatus(stream, track, filters, NOTE_ON, self.note, self.velocity)
		if track.running_status == (NOTE_OFF + self.channel):
			self.WriteRunningStatus(stream, track, filters, NOTE_ON, self.note, self.velocity)
		else:
			self.WriteRunningStatus(stream, track, filters, NOTE_ON, self.note, 0)
	def __str__ (self):
		return '%s: ch=%d n=%d v=%d' % (self.name, self.channel, self.note, self.velocity)

#---------------------------------------------------------------
# NoteOnEvent
#---------------------------------------------------------------
class NoteOnEvent (MIDIEvent):
	def __init__ (self, ticks, seq, channel, note, velocity, note_length, note_off_velocity):
		self.name = 'NoteOn'
		self.msg_type = NOTE_ON
		self.ticks = ticks
		self.seq = seq
		self.channel = channel
		self.note = note
		self.velocity = velocity
		self.note_length = note_length
		self.note_off_velocity = note_off_velocity
	@staticmethod
	def ReadFromStream (stream, seq, ticks, msg_type):
		channel = msg_type & 0x0f
		note = ReadByte(stream)
		velocity = ReadByte(stream)
		if msg_type & 0xf0 != NOTE_ON:
			stream.seek(-2,1)
			raise MIDIFileException(stream, MSG_TYPE_MISMATCH)
		if velocity == 0:
			return NoteOffEvent(ticks, seq, channel, note, velocity)
		return NoteOnEvent(ticks, seq, channel, note, velocity, None, None)
	def WriteToStream (self, stream, track, filters=None):
		self.WriteRunningStatus(stream, track, filters, NOTE_ON, self.note, self.velocity)
	def __str__ (self):
		if self.note_length is not None:
			return '%s: ch=%d n=%d v=%d l=%d' % (self.name, self.channel, self.note, self.velocity, self.note_length)
		else:
			return '%s: ch=%d n=%d v=%d' % (self.name, self.channel, self.note, self.velocity)

#---------------------------------------------------------------
# PolyKeyPressureEvent
#---------------------------------------------------------------
class PolyKeyPressureEvent (MIDIEvent):
	def __init__ (self, ticks, seq, channel, note, value):
		self.name = 'PolyKeyPressure'
		self.msg_type = POLY_KEY_PRESSURE
		self.ticks = ticks
		self.seq = seq
		self.channel = channel
		self.note = note
		self.value = value
	@staticmethod
	def ReadFromStream (stream, seq, ticks, msg_type):
		channel = msg_type & 0x0f
		note = ReadByte(stream)
		value = ReadByte(stream)
		if msg_type & 0xf0 != POLY_KEY_PRESSURE:
			stream.seek(-2,1)
			raise MIDIFileException(stream, MSG_TYPE_MISMATCH)
		return PolyKeyPressureEvent(ticks, seq, channel, note, value)
	def WriteToStream (self, stream, track, filters=None):
		self.WriteRunningStatus(stream, track, filters, POLY_KEY_PRESSURE, self.note, self.value)
	def __str__ (self):
		return '%s: ch=%d n=%d v=%d' % (self.name, self.channel, self.note, self.value)

#---------------------------------------------------------------
# ControlChangeEvent
#---------------------------------------------------------------
class ControlChangeEvent (MIDIEvent):
	def __init__ (self, ticks, seq, channel, controller, value):
		self.name = 'ControlChange'
		self.msg_type = CONTROL_CHANGE
		self.ticks = ticks
		self.seq = seq
		self.channel = channel
		self.controller = controller
		self.value = value
	@staticmethod
	def ReadFromStream (stream, seq, ticks, msg_type):
		channel = msg_type & 0x0f
		controller = ReadByte(stream)
		value = ReadByte(stream)
		if msg_type & 0xf0 != CONTROL_CHANGE:
			stream.seek(-2,1)
			raise MIDIFileException(stream, MSG_TYPE_MISMATCH)
		if controller >= 120:
			return ChannelModeEvent(ticks, seq, channel, controller, value)
		return ControlChangeEvent(ticks, seq, channel, controller, value)
	def WriteToStream (self, stream, track, filters=None):
		self.WriteRunningStatus(stream, track, filters, CONTROL_CHANGE, self.controller, self.value)
	def __str__ (self):
		return '%s: ch=%d c=%d v=%d' % (self.name, self.channel, self.controller, self.value)

#---------------------------------------------------------------
# ChannelModeEvent
#---------------------------------------------------------------
class ChannelModeEvent (MIDIEvent):
	def __init__ (self, ticks, seq, channel, controller, value):
		self.name = 'ChannelMode'
		self.msg_type = CONTROL_CHANGE
		self.ticks = ticks
		self.seq = seq
		self.channel = channel
		self.controller = controller
		self.value = value
	@staticmethod
	def ReadFromStream (stream, seq, ticks, msg_type):
		channel = msg_type & 0x0f
		controller = ReadByte(stream)
		value = ReadByte(stream)
		if msg_type & 0xf0 != CONTROL_CHANGE:
			stream.seek(-2,1)
			raise MIDIFileException(stream, MSG_TYPE_MISMATCH)
		if  controller < 120:
			return ControlChangeEvent(ticks, seq, channel, controller, value)
		return ChannelModeEvent(ticks, seq, channel, value)
	def WriteToStream (self, stream, track, filters=None):
		self.WriteRunningStatus(stream, track, filters, CONTROL_CHANGE, self.controller, self.value)
	def __str__ (self):
		return '%s: ch=%d c=%d v=%d' % (self.name, self.channel, self.controller, self.value)

#---------------------------------------------------------------
# ProgramChangeEvent
#---------------------------------------------------------------
class ProgramChangeEvent (MIDIEvent):
	def __init__ (self, ticks, seq, channel, program):
		self.name = 'ProgramChange'
		self.msg_type = PROGRAM_CHANGE
		self.ticks = ticks
		self.seq = seq
		self.channel = channel
		self.program = program
	@staticmethod
	def ReadFromStream (stream, seq, ticks, msg_type):
		channel = msg_type & 0x0f
		program = ReadByte(stream)
		if msg_type & 0xf0 != PROGRAM_CHANGE:
			stream.seek(-1,1)
			raise MIDIFileException(stream, MSG_TYPE_MISMATCH)
		return ProgramChangeEvent(ticks, seq, channel, program)
	def WriteToStream (self, stream, track, filters=None):
		self.WriteRunningStatus(stream, track, filters, PROGRAM_CHANGE, self.program)
	def __str__ (self):
		return '%s: ch=%d p=%d' % (self.name, self.channel, self.program)

#---------------------------------------------------------------
# ChannelPressureEvent
#---------------------------------------------------------------
class ChannelPressureEvent (MIDIEvent):
	def __init__ (self, ticks, seq, channel, value):
		self.name = 'ChannelPressure'
		self.msg_type = CHANNEL_PRESSURE
		self.ticks = ticks
		self.seq = seq
		self.channel = channel
		self.value = value
	@staticmethod
	def ReadFromStream (stream, seq, ticks, msg_type):
		channel = msg_type & 0x0f
		value = ReadByte(stream)
		if msg_type & 0xf0 != CHANNEL_PRESSURE:
			stream.seek(-1,1)
			raise MIDIFileException(stream, MSG_TYPE_MISMATCH)
		return ChannelPressureEvent(ticks, seq, channel, value)
	def WriteToStream (self, stream, track, filters=None):
		self.WriteRunningStatus(stream, track, filters, CHANNEL_PRESSURE, self.value)
	def __str__ (self):
		return '%s: ch=%d v=%d' % (self.name, self.channel, self.value)

#---------------------------------------------------------------
# PitchBendEvent
#---------------------------------------------------------------
class PitchBendEvent (MIDIEvent):
	def __init__ (self, ticks, seq, channel, value):
		self.name = 'PitchBend'
		self.msg_type = PITCH_BEND
		self.ticks = ticks
		self.seq = seq
		self.channel = channel
		self.value = value
	@staticmethod
	def ReadFromStream (stream, seq, ticks, msg_type):
		channel = msg_type & 0x0f
		value = (ReadByte(stream) << 7) + ReadByte(stream) - 0x2000
		if msg_type & 0xf0 != PITCH_BEND:
			stream.seek(-2,1)
			raise MIDIFileException(stream, MSG_TYPE_MISMATCH)
		return PitchBendEvent(ticks, seq, channel, value)
	def WriteToStream (self, stream, track, filters=None):
		value = self.value + 0x2000
		if value < 0:
			value = 0
		if value > 0x3fff:
			value = 0x3fff
		self.WriteRunningStatus(stream, track, filters, PITCH_BEND, value >> 7, value & 0x7f)
	def __str__ (self):
		return '%s: ch=%d v=%d' % (self.name, self.channel, self.value)

#---------------------------------------------------------------
# SysExEvent
#---------------------------------------------------------------
class SysExEvent (MIDIEvent):
	def __init__ (self, ticks, seq, msg):
		self.name = 'SysEx'
		self.msg_type = SYSEX
		self.ticks = ticks
		self.seq = seq
		self.length = len(msg)
		self.msg = msg
	@staticmethod
	def ReadFromStream (stream, seq, ticks, msg_type):
		pos = stream.tell()
		length = ReadVarLenQty(stream)
		msg = ReadBytes(stream, length)
		if msg_type != SYSEX:
			stream.seek(pos,0)
			raise MIDIFileException(stream, MSG_TYPE_MISMATCH)
		return SysExEvent(ticks, seq, msg)
	def WriteToStream (self, stream, track, filters=None):
		if not self.CheckFilters(filters):
			return
		self.WriteTicks(stream, track)
		WriteByte(stream, SYSEX)
		WriteVarLenQty(stream, self.length)
		WriteBytes(stream, self.msg)
		track.running_status = None
	def __str__ (self):
		fmt_str = '%s: f0' + ' %02x'*self.length
		return fmt_str % ((self.name,) + tuple(self.msg))

#---------------------------------------------------------------
# SysExContEvent
#---------------------------------------------------------------
class SysExContEvent (MIDIEvent):
	def __init__ (self, ticks, seq, msg):
		self.name = 'SysEx+'
		self.msg_type = END_SYSEX
		self.ticks = ticks
		self.seq = seq
		self.length = len(msg)
		self.msg = msg
	@staticmethod
	def ReadFromStream (stream, seq, ticks, msg_type):
		pos = stream.tell()
		length = ReadVarLenQty(stream)
		msg = ReadBytes(stream, length)
		if msg_type != END_SYSEX:
			stream.seek(pos,0)
			raise MIDIFileException(stream, MSG_TYPE_MISMATCH)
		return SysExContEvent(ticks, seq, msg)
	def WriteToStream (self, stream, track, filters=None):
		if not self.CheckFilters(filters):
			return
		self.WriteTicks(stream, track)
		WriteByte(stream, END_SYSEX)
		WriteVarLenQty(stream, self.length)
		WriteBytes(stream, self.msg)
		track.running_status = None
	def __str__ (self):
		fmt_str = '%s:' + ' %02x'*self.length
		return fmt_str % ((self.name,) + tuple(self.msg))

#---------------------------------------------------------------
# MetaEvent
#---------------------------------------------------------------
class MetaEvent (MIDIEvent):
	def __init__ (self, ticks, seq, meta_type, msg):
		self.name = 'MetaEvent'
		self.msg_type = META_EVENT
		self.ticks = ticks
		self.seq = seq
		self.meta_type = meta_type
		self.length = len(msg)
		self.msg = msg
	@staticmethod
	def ReadFromStream (stream, seq, ticks, msg_type):
		pos = stream.tell()
		meta_type = ReadByte(stream)
		length = ReadVarLenQty(stream)
		msg = ReadBytes(stream, length)
		if msg_type != META_EVENT:
			stream.seek(pos,0)
			raise MIDIFileException(stream, MSG_TYPE_MISMATCH)
		obj = MetaEvent(ticks, seq, meta_type, msg)
		return obj
	def WriteToStream (self, stream, track, filters=None):
		if not self.CheckFilters(filters):
			return
		self.WriteTicks(stream, track)
		WriteByte(stream, META_EVENT)
		WriteByte(stream, self.meta_type)
		WriteVarLenQty(stream, self.length)
		WriteBytes(stream, self.msg)
		track.running_status = None
	def __str__ (self):
		fmt_str = '%s: %02x' + ' %02x'*self.length
		return fmt_str % ((self.name, self.meta_type) + tuple(self.msg))

#---------------------------------------------------------------
# MIDIControllers
#---------------------------------------------------------------
class MIDIControllers (object):
	def __init__ (self):
		self.controllers = []
		self.rpns = []
		for channel in range(16):
			self.controllers.append({})
			self.controllers[channel] = copy.deepcopy(DEFAULT_CONTROLLER_VALUES)
			self.rpns.append({})
			self.rpns[channel] = copy.deepcopy(DEFAULT_RPN_VALUES)
		self.pitchbend = [0] * 16
		self.program = [-1] * 16
		self.pressure = [0] * 16

	def __str__ (self):
		output = []
		for channel in range(16):
			output.append('channel=%d' % channel)
			output.append('  program=%d' % self.program[channel])
			output.append('  pressure=%d' % self.pressure[channel])

			output.append('  controllers')
			for controller in self.controllers[channel].keys():
				output.append('    %03d: %03d' % (controller, self.controllers[channel][controller]))

			output.append('  rpns')
			for rpn in self.rpns[channel].keys():
				output.append('    %05d: %05d>' % (controller, self.rpns[channel][rpn]))
		return '\n'.join(output)


	def Event (self, event):
		"""Process an event and save any changes in controller values"""
		# process control changes
		if event.msg_type == CONTROL_CHANGE:
			self.ControlChange(event)
		elif event.msg_type == CHANNEL_PRESSURE:
			self.PressureChange(event)
		elif event.msg_type == PROGRAM_CHANGE:
			self.ProgramChange(event)
		elif event.msg_type == PITCH_BEND:
			self.PitchBendChange(event)

	def PitchBendChange (self, event):
		"""Monitor pitch bend change."""
		self.pitchbend[event.channel] = event.value

	def ProgramChange (self, event):
		"""Monitor program change."""
		self.program[event.channel] = event.program

	def ControlChange (self, event):
		"""Monitor control change."""
		controller = event.controller
		if controller in MONITOR_CONTROLLERS:
			channel = event.channel
			self.controllers[channel][controller] = event.value
			if (controller == CTRL_RPN_DATA_MSB) or (controller == CTRL_RPN_DATA_LSB):
				rpn = (self.controllers[channel][CTRL_RPN_MSB] << 7) + self.controllers[channel][CTRL_RPN_LSB]
				if rpn in MONITOR_RPNS:
					value = (self.controllers[channel][CTRL_RPN_DATA_MSB] << 7) + self.controllers[channel][CTRL_RPN_DATA_LSB]
					self.rpns[channel][rpn] = value

		# reset controllers
		elif event.controller == CTRL_RESET_CONTROLLERS:
			self.ResetControllers[event.channel]

	def PressureChange (self, event):
		"""Monitor pressure change."""
		self.pressure[event.channel] = event.value

	def ResetControllers (self, channel):
		"""Reset controllers to default."""
		self.controllers[channel] = DEFAULT_CONTROLLER_VALUES
		self.rpns[channel] = DEFAULT_RPN_VALUES
		self.pressure[channel] = 0

	def GenerateEventList (self, ticks, ref_values=None):
		"""Generate an event list based on controller differences."""
		events = EventList()

		# if no reference values, based on default values
		if ref_values is None:
			ref_values = MIDIControllers()

		# iterate through 16 MIDI channels
		for channel in range(16):

			# generate RPN changes
			for rpn in self.rpns[channel].keys():
				value = self.rpns[channel][rpn]
				if value != ref_values.rpns[channel][rpn]:
					events.append(ControlChangeEvent(ticks, -1, channel, CTRL_RPN_MSB, rpn >> 7))
					events.append(ControlChangeEvent(ticks, -1, channel, CTRL_RPN_LSB, rpn & 0x7f))
					events.append(ControlChangeEvent(ticks, -1, channel, CTRL_RPN_DATA_MSB, value >> 7))
					events.append(ControlChangeEvent(ticks, -1, channel, CTRL_RPN_DATA_LSB, value & 0x7f))

			# generate controller changes
			for controller in self.controllers[channel].keys():
				if self.controllers[channel][controller] != ref_values.controllers[channel][controller]:
					events.append(ControlChangeEvent(ticks, -1, channel, controller, self.controllers[channel][controller]))

			# generate pressure changes
			if self.pressure[channel] != ref_values.pressure[channel]:
				events.append(ChannelPressureEvent(ticks, -1, channel, self.pressure[channel]))

			# generate program changes
			if self.program[channel] != ref_values.program[channel]:
				if self.program[channel] in range(128):
					events.append(ProgramChangeEvent(ticks, -1, channel, self.program[channel]))

			# generate pitch bend changes
			if self.pitchbend[channel] != ref_values.pitchbend[channel]:
				if self.pitchbend[channel] in range(-8192,8191):
					events.append(PitchBendEvent(ticks, -1, channel, self.pitchbend[channel]))

		return events

#---------------------------------------------------------------
# EventList
#---------------------------------------------------------------
class EventList (list):
	def __init__ (self):
		list.__init__(self)

	def FixNoteLengths (self):
		midi_file_logger.debug('Fix note lengths')

		# search for note-on's in event list
		for index in range(len(self)):
			event = self[index]
			if event.msg_type == NOTE_ON:
				note_off_ticks = event.ticks + event.note_length

				# check for note-on occuring before end of current note
				for i in range(index + 1, len(self)):
					event_to_check = self[i]
					if event_to_check.ticks >= note_off_ticks:
						break

					# adjust note length
					if (event_to_check.msg_type == NOTE_ON) and (event_to_check.note == event.note):
						midi_file_logger.debug('Adjusting note length @ %d' % event.ticks)
						event.note_length = event_to_check.ticks - event.ticks
						break

	def ChaseControllers (self, end_seq, start_seq = 0, values = None):
		midi_file_logger.debug('ChaseControllers from %d to %d' % (start_seq, end_seq))

		# initialize controller values
		if values is None:
			values = MIDIControllers()

		# chase controllers in track
		for i in range(start_seq, min(end_seq, len(self))):
			values.Event(self[i])

		# return new values
		return values
		
	def SelectEvents (self, start, end):
		midi_file_logger.debug('SelectEvents: %d to %d' % (start, end))
		selected = EventList()
		for event in self:
			if event.ticks >= start:
				if event.ticks >= end:
					break
				midi_file_logger.debug('SelectEvent: %s' % event.__str__())
				selected.append(event)
		return selected

	def MergeEvents (self, events):
		# copy events and sort them by ticks/sequence#
		self.extend(events)
		self.SortEvents()
		
	def InsertEvents (self, events, seq):
		self[seq:seq] = events
		self.RenumberSeq()

	def DeleteEvents (self, start_index, end_index, move_meta_events=None):
		# default parameters
		if start_index is None:
			start_index = 0
		if end_index is None:
			end_index = len(self)

		#print("\n")
		#for evt in self[start_index:end_index]:
		#	print("%d %s" % (evt.ticks, evt))

		# delete events
		delete_count = 0
		move_count = 0
		for event in self[start_index:end_index]:
			#Bth; Added this so we always get clip end events; clips that ended on last measure wouldn't end on repeat
			if (event.msg_type == CONTROL_CHANGE) and \
			        (event.controller == JET_EVENT_TRIGGER_CLIP) and \
			        ((event.value & 0x40) != 0x40):
				pass
			else:
				if (move_meta_events is None) or (event.msg_type != META_EVENT):
					self.remove(event)
					delete_count += 1
					
				# move meta-events
				else:
					event.ticks = move_meta_events
					move_count += 1
				
		midi_file_logger.debug('DeleteEvents: deleted %d events in range(%s:%s)' % (delete_count, start_index, end_index))
		midi_file_logger.debug('DeleteEvents: moved %d events in range(%s:%s)' % (move_count, start_index, end_index))

			
	def SeekEvent (self, pos):
		for i in range(len(self)):
			if self[i].ticks >= pos:
				return i
		return None

	def RenumberSeq (self):
		seq = 0
		for event in self:
			event.seq = seq
			seq += 1

	def SortEvents (self):
		self.sort(self.EventSorter)
		self.RenumberSeq()

	@staticmethod
	def EventSorter (x, y):
		if x.ticks == y.ticks:
			return cmp(x.seq, y.seq)
		else:
			return cmp(x.ticks, y.ticks)

	def DumpEvents (self, output, timebase):
		if output is not None:
			for event in self:
				output.write('%s\n' % event.TimeEventStr(timebase))
		else:
			for event in self:
				midi_file_logger.debug(event.TimeEventStr(timebase))

#---------------------------------------------------------------
# MIDITrack
#---------------------------------------------------------------
class MIDITrack (object):
	"""The MIDITrack class implements methods for reading, parsing,
	modifying, and writing tracks in Standard MIDI Files (SMF).

	"""
	def __init__ (self):
		self.length = 0
		self.events = EventList()
		self.end_of_track = None
		self.channel = None
		self.name = None
	
	def ReadFromStream (self, stream, offset, file_size):
		self.stream = stream
		ticks = 0
		seq = 0
		running_status = None
		tick_warning_level = stream.timebase.ppqn * LARGE_TICK_WARNING
		
		# read the track header - verify it's an SMF track
		stream.seek(offset)
		bytes = stream.read(struct.calcsize(SMF_TRACK_HEADER_FMT))
		riff_tag, track_len = struct.unpack(SMF_TRACK_HEADER_FMT, bytes)
		midi_file_logger.debug('SMF track header\n  Tag:      %s\n  TrackLen: %d' % (riff_tag, track_len))
		if (riff_tag != SMF_TRACK_RIFF_TAG):
			raise MIDIFileException(stream, MSG_INVALID_TRACK_HEADER)
		self.start = stream.tell()
		
		# check for valid track length
		if (self.start + track_len) > file_size:
			stream.Warning('Ignoring illegal track length - %d exceeds length of file' % track_len)
			track_len = None
			
		# read the entire track
		note_on_list = []
		while 1:

			# save current position
			pos = stream.tell()

			# check for end of track
			if track_len is not None:
				if (pos - self.start) >= track_len:
					break

			# are we past end of track?
			if self.end_of_track:
				stream.Warning('Ignoring data encountered beyond end-of-track meta-event')
				break;
		
			# read delta timestamp
			delta = ReadVarLenQty(stream)
			if ticks > tick_warning_level:
				stream.Warning('Tick value is excessive - possibly corrupt data?')
			ticks += delta
				
			# get the event type and process it
			msg_type = ReadByte(stream)

			# if data byte, check for running status
			if msg_type & 0x80 == 0:

				# use running status
				msg_type = running_status

				# back up so event can process data
				stream.seek(-1,1)

				# if no running status, we have a problem
				if not running_status:
					stream.Warning('Ignoring data byte received with no running status')

			# create event type from stream
			event = MIDIEvent.ReadFromStream(stream, seq, ticks, msg_type)
			
			if self.channel == None:
				try:
					self.channel = event.channel
				except AttributeError:
					pass
					
			# track note-ons
			if event.msg_type == NOTE_ON:

				"""
				Experimental code to clean up overlapping notes
				Clean up now occurs during write process
				
				for note_on in note_on_list:
					if (event.channel == note_on.channel) and (event.note == note_on.note):
						stream.Warning('Duplicate note-on\'s encountered without intervening note-off')
						stream.Warning('  [%s]: %s' % (stream.timebase.ConvertTicksToStr(event.ticks), event.__str__()))
						note_on.note_length = event.ticks - note_on.ticks - 1
						if note_on.note_length <= 0:
							stream.Warning('Eliminating duplicate note-on')
							event.ticks = note_on.ticks
							self.events.remove(note_on)
				"""
				
				note_on_list.append(event)

			# process note-offs
			if event.msg_type == NOTE_OFF:
				for note_on in note_on_list[:]:
					if (event.channel == note_on.channel) and (event.note == note_on.note):
						note_on.note_length = event.ticks - note_on.ticks
						note_on.note_off_velocity = event.velocity
						note_on_list.remove(note_on)
						break
				#else:
				#	stream.Warning('Note-off encountered without corresponding note-on')
				#	stream.Warning('  [%s]: %s' % (stream.timebase.ConvertTicksToStr(event.ticks), event.__str__()))

			# check for end of track
			elif event.msg_type == META_EVENT and event.meta_type == META_EVENT_END_OF_TRACK:
				self.end_of_track = event.ticks

			# BTH; get track name
			elif event.msg_type == META_EVENT and event.meta_type == META_EVENT_SEQUENCE_TRACK_NAME:
				self.name = array.array('B', event.msg).tostring()
				
			# append event to event list
			else:
				self.events.append(event)
				seq += 1

			# save position for port-mortem
			stream.last_good_event = pos

			# update running statusc_str(
			if msg_type < 0xf0:
				running_status = msg_type
			elif (msg_type < 0xf8) or (msg_type == 0xff):
				running_status = None

		# check for stuck notes
		#if len(note_on_list):
		#	stream.Warning('Note-ons encountered without corresponding note-offs')

		# check for missing end-of-track meta-event
		if self.end_of_track is None:
			self.last_tick = self.events[-1].ticks
			stream.Warning('End of track encountered with no end-of-track meta-event')

		# if track length was bad, correct it
		if track_len is None:
			track_len = stream.tell() - offset - 8

		return track_len

	def Write (self, stream, filters=None):
		# save current file position so we can write header
		header_loc = stream.tell()
		stream.seek(header_loc + struct.calcsize(SMF_TRACK_HEADER_FMT))

		# save a copy of the event list so we can restore it
		save_events = copy.copy(self.events)

		# create note-off events
		index = 0
		while 1:
			if index >= len(self.events):
				break

			# if note-on event, create a note-off event
			event = self.events[index]
			index += 1
			if event.msg_type == NOTE_ON:
				note_off = NoteOffEvent(event.ticks + event.note_length, index, event.channel, event.note, event.note_off_velocity)

				# insert note-off in list
				for i in range(index, len(self.events)):
					if self.events[i].ticks >= note_off.ticks:
						self.events.insert(i, note_off)
						break
				else:
					self.events.append(note_off)

		# renumber list
		self.events.RenumberSeq()

		# write the events
		self.running_status = None
		self.ticks = 0
		for event in self.events:

			# write event
			event.WriteToStream(stream, self, filters)

		# restore original list (without note-off events)
		self.events = save_events

		# write the end-of-track meta-event
		MetaEvent(self.end_of_track, 0, META_EVENT_END_OF_TRACK,[]).WriteToStream(stream, self, None)

		# write track header
		end_of_track = stream.tell()
		track_len = end_of_track - header_loc - struct.calcsize(SMF_TRACK_HEADER_FMT)
		stream.seek(header_loc)
		bytes = struct.pack(SMF_TRACK_HEADER_FMT, SMF_TRACK_RIFF_TAG, track_len)
		stream.write(bytes)
		stream.seek(end_of_track)

	def Trim (self, start, end, slide=True, chase_controllers=True, delete_meta_events=False, quantize=0):
		controllers = None

		if quantize:
			# quantize events just before start
			for event in self.events.SelectEvents(start - quantize, start):
				midi_file_logger.debug('Trim: Moving event %s to %d' % (event.__str__(), start))
				event.ticks = start

			# quantize events just before end
			for event in self.events.SelectEvents(end - quantize, end):
				midi_file_logger.debug('Trim: Moving event %s to %d' % (event.__str__(), end))
				event.ticks = end

		# trim start
		if start:

			# find first event inside trim
			start_event = self.events.SeekEvent(start)
			if start_event is not None:

				# chase controllers to cut point
				if chase_controllers:
					controllers = self.events.ChaseControllers(self.events[start_event].seq)
					controller_events = controllers.GenerateEventList(0)
					midi_file_logger.debug('Trim: insert new controller events at %d:' % start)
					controller_events.DumpEvents(None, self.stream.timebase)
					self.events.InsertEvents(controller_events, start_event)

				# delete events					
				midi_file_logger.debug('Trim: deleting events up to event %d' % start_event)
				if delete_meta_events:
					self.events.DeleteEvents(None, start_event, None)
				else:
					self.events.DeleteEvents(None, start_event, start)

			# delete everything except metadata
			else:
				self.events.DeleteEvents(None, None, start)

		# trim end
		end_event = self.events.SeekEvent(end)
		if end_event is not None:
			midi_file_logger.debug('Trim: trimming section starting at event %d' % end_event)
			self.events.DeleteEvents(end_event, None)

		# trim any notes that extend past the end
		for event in self.events:
			if event.msg_type == NOTE_ON:
				if (event.ticks + event.note_length) > end:
					midi_file_logger.debug('Trim: trimming note that extends past end %s' % event.TimeEventStr(self.stream.timebase))
					event.note_length = end - event.ticks
					if event.note_length <= 0:
						raise 'Error in note length - note should have been deleted'

		midi_file_logger.debug('Trim: initial end-of-track: %d' % self.end_of_track)
		self.end_of_track = min(self.end_of_track, end)

		# slide events to start of track to fill hole
		if slide and start:
			midi_file_logger.debug('Trim: sliding events: %d' % start)
			for event in self.events:
				if event.ticks > start:
					event.ticks -= start
				else:
					event.ticks = 0
			self.end_of_track = max(0, self.end_of_track - start)
		midi_file_logger.debug('Trim: new end-of-track: %d' % self.end_of_track)

		self.events.RenumberSeq()
		self.events.FixNoteLengths()

	def DumpEvents (self, output):
		self.events.DumpEvents(output, self.stream.timebase)
		if output is not None:
			output.write('[%s]: end-of-track\n' % self.stream.timebase.ConvertTicksToStr(self.end_of_track))
		else:
			midi_file_logger.debug('[%s]: end-of-track' % self.stream.timebase.ConvertTicksToStr(self.end_of_track))


#---------------------------------------------------------------
# MIDIFile
#---------------------------------------------------------------
class MIDIFile (file):
	"""The MIDIFile class implements methods for reading, parsing,
	modifying, and writing Standard MIDI Files (SMF).

	"""
	def __init__ (self, name, mode):
		file.__init__(self, name, mode)
		self.timebase = TimeBase()

	def ReadFromStream (self, start_offset=0, file_size=None):
		"""Parse the MIDI file creating a list of properties, tracks,
		and events based on the contents of the file.

		"""

		# determine file size - without using os.stat
		if file_size == None:
			self.start_offset = start_offset
			self.seek(0,2)
			file_size = self.tell() - self.start_offset
			self.seek(start_offset,0)
		else:
			file_size = file_size

		# for error recovery
		self.last_good_event = None
		self.error_loc = None

		# read the file header - verify it's an SMF file
		bytes = self.read(struct.calcsize(SMF_HEADER_FMT))
		riff_tag, self.hdr_len, self.format, self.num_tracks, self.timebase.ppqn = struct.unpack(SMF_HEADER_FMT, bytes)
		midi_file_logger.debug('SMF header\n  Tag:       %s\n  HeaderLen: %d\n  Format:    %d\n  NumTracks: %d\n  PPQN:      %d\n' % \
			(riff_tag, self.hdr_len, self.format, self.num_tracks, self.timebase.ppqn))

		# sanity check on header
		if (riff_tag != SMF_RIFF_TAG) or (self.format not in range(2)):
			raise MIDIFileException(self, MSG_NOT_SMF_FILE)

		# check for odd header size
		if self.hdr_len + 8 != struct.calcsize(SMF_HEADER_FMT):
			self.Warning('SMF file has unusual header size: %d bytes' % self.hdr_len)

		# read each of the tracks
		offset = start_offset + self.hdr_len + 8
		self.tracks = []
		self.end_of_file = 0
		for i in range(self.num_tracks):
			#print("Track: %d" % i)

			# parse the track
			track = MIDITrack()
			length = track.ReadFromStream(self, offset, file_size)
			track.trackNum = i
			
			self.tracks.append(track)

			# calculate offset to next track
			offset += length + 8

			# determine time of last event
			self.end_of_file = max(self.end_of_file, track.end_of_track)

		# if start_offset is zero, the final offset should match the file length
		if (offset - start_offset) != file_size:
			self.Warning('SMF file size is incorrect - should be %d, was %d' % (file_size, offset))
		
	def Save (self, offset=0, filters=None):
		"""Save this file back to disk with modifications."""
		if (not 'w' in self.mode) and (not '+' in self.mode):
			raise MIDIFileException(self, 'Cannot write to file in read-only mode')
		self.Write(self, offset, filters)

	def SaveAs (self, filename, offset=0, filters=None):
		"""Save MIDI data to new file."""
		output_file = MIDIFile(filename, 'wb')
		self.Write(output_file, offset, filters)
		output_file.close()

	def Write (self, output_file, offset=0, filters=None):
		"""This function does the actual work of writing the file."""
		# write the file header
		output_file.seek(offset)
		bytes = struct.pack(SMF_HEADER_FMT, SMF_RIFF_TAG, struct.calcsize(SMF_HEADER_FMT) - 8, self.format, self.num_tracks, self.timebase.ppqn)
		output_file.write(bytes)

		# write out the tracks
		for track in self.tracks:
			track.Write(output_file, filters)

		# flush the data to disk
		output_file.flush()

	def ConvertToType0 (self):
		"""Convert a file to type 0."""
		if self.format == 0:
			midi_file_logger.warning('File is already type 0 - ignoring request to convert')
			return

		# convert to type 0
		for track in self.tracks[1:]:
			self.tracks[0].MergeEvents(track.events)
		self.tracks = self.tracks[:1]
		self.num_tracks = 1
		self.format = 0

	def DeleteEmptyTracks (self):
		"""Delete any tracks that do not contain MIDI messages"""
		track_num = 0
		for track in self.tracks[:]:
			for event in self.tracks.events:
				if event.msg_type in MIDI_MESSAGES:
					break;
				else:
					midi_file_logger.debug('Deleting track %d' % track_num)
					self.tracks.remove(track)
			track_num += 1

	def ConvertToTicks (self, measures, beats, ticks):
		return self.timebase.ConvertToTicks(measures, beats, ticks)

	def Trim (self, start, end, quantize=0, chase_controllers=True):
		track_num = 0
		for track in self.tracks:
			midi_file_logger.debug('Trimming track %d' % track_num)
			track.Trim(start, end, quantize=quantize, chase_controllers=chase_controllers)
			track_num += 1

	def DumpTracks (self, output=None):
		track_num = 0
		for track in self.tracks:
			if output is None:
				midi_file_logger.debug('*** Track %d ***' % track_num)
			else:
				output.write('*** Track %d ***' % track_num)
			track.DumpEvents(output)
			track_num += 1

	def Warning (self, msg):
		midi_file_logger.warning('[%d]: %s' % (self.tell(), msg))

	def Error (self, msg):
		midi_file_logger.error('[%d]: %s' % (self.tell(), msg))

	def DumpError (self):
		if self.last_good_event:
			midi_file_logger.error('Dumping from last good event:')
			pos = self.last_good_event - 16
			length = self.error_loc - pos + 16
		elif self.error_loc:
			midi_file_logger.error('Dumping from 16 bytes prior to error:')
			pos = self.error_loc
			length = 32
		else:
			midi_file_logger.error('No dump information available')
			return

		self.seek(pos, 0)
		for i in range(length):
			if i % 16 == 0:
				if i:
					midi_file_logger.error(' '.join(debug_out))
				debug_out = ['%08x:' % (pos + i)]
			byte = self.read(1)
			if len(byte) == 0:
				break;
			debug_out.append('%02x' % ord(byte))
		if i % 16 > 0:
			midi_file_logger.error(' '.join(debug_out))

def GetMidiInfo(midiFile):
	"""Bth; Get MIDI info"""
	
	class midiData(object):
		def __init__ (self):
			self.err = 1
			self.endMbt = "0:0:0"
			self.totalTicks = 0
			self.maxTracks = 0
			self.maxMeasures = 0
			self.maxBeats = 0
			self.maxTicks = 0
			self.totalTicks = 0
			self.timebase = None
			self.ppqn = 0
			self.beats_per_measure = 0
			self.trackList = []
			
	md = midiData()
	
	try:
		m = MIDIFile(midiFile, 'rb')
		m.ReadFromStream()
		
		for track in m.tracks:
			if track.channel is not None:
				empty = False 
				trk = track.channel + 1
			else:
				empty = True	
				trk = ''		
			md.trackList.append(trackGrid(track.trackNum, trk, track.name, empty))
				
		md.endMbt = m.timebase.ConvertTicksToMBT(m.end_of_file)
		md.endMbtStr = "%d:%d:%d" % (md.endMbt[0], md.endMbt[1], md.endMbt[2])
		md.maxMeasures = md.endMbt[0]
		md.maxBeats = 4
		md.maxTicks = m.timebase.ppqn
		md.maxTracks = m.num_tracks
		md.totalTicks = m.end_of_file
		md.timebase = m.timebase
		md.ppqn = m.timebase.ppqn
		md.beats_per_measure = m.timebase.beats_per_measure

		#add above if more added
		md.err = 0
		
		m.close()
	except:
		raise
		pass
	
	return md
	
		


#---------------------------------------------------------------
# main
#---------------------------------------------------------------
if __name__ == '__main__':
	sys = __import__('sys')
	os = __import__('os')

	# initialize root logger
	root_logger = logging.getLogger('')
	root_logger.setLevel(logging.NOTSET)

	# initialize console handler
	console_handler = logging.StreamHandler()
	console_handler.setFormatter(logging.Formatter('%(message)s'))
	console_handler.setLevel(logging.DEBUG)
	root_logger.addHandler(console_handler)

	files = []
	dirs = []
	last_arg = None
	sysex_filter = False
	drum_filter = False
	convert = False

	# process args
	for arg in sys.argv[1:]:

		# previous argument implies this argument
		if last_arg is not None:
			if last_arg == '-DIR':
				dirs.append(arg)
				last_arg = None

		# check for switch
		elif arg[0] == '-':
			if arg == '-DIR':
				last_arg = arg
			elif arg == '-SYSEX':
				sysex_filter = True
			elif arg == '-DRUMS':
				drum_filter = True
			elif arg == '-CONVERT':
				convert = True
			else:
				midi_file_logger.error('Bad option %s' % arg)

		# must be a filename
		else:
			files.append(arg)

	# setup filters
	filters = []
	if sysex_filter:
		filters.append(EventTypeFilter((SYSEX,)))
	if drum_filter:
		filters.append(ChannelFilter((9,),False))
		

	# process dirs
	for d in dirs:
		for root, dir_list, file_list in os.walk(d):
			for f in file_list:
				if f.endswith('.mid'):
					files.append(os.path.join(root, f))

	# process files
	bad_files = []
	for f in files:
		midi_file_logger.info('Processing file %s' % f)
		midiFile = MIDIFile(f, 'rb')
		try:
			midiFile.ReadFromStream()
			
			#midiFile.DumpTracks()
			#print('[%s]: end-of-track\n' % midiFile.timebase.ConvertTicksToStr(midiFile.end_of_file))

			# convert to type 0
			if convert and (midiFile.format == 1):
				midiFile.Convert(0)
				converted = True
			else:
				converted = False

			# write processed file
			if converted or len(filters):
				midiFile.SaveAs(f[:-4] + '-mod.mid', filters)
				
		except MIDIFileException, X:
			bad_files.append(f)
			midi_file_logger.error('Error in file %s' % f)
			midi_file_logger.error(X)
			midiFile.DumpError()
		midiFile.close()

	# dump problem files
	if len(bad_files):
		midi_file_logger.info('The following file(s) had errors:')
		for f in bad_files:
			midi_file_logger.info(f)
	else:
		midi_file_logger.info('All files read successfully')




"""
 File:  
 JetCreator.py
 
 Contents and purpose:
 Jet file creation utility for JET sound engine
 
 Copyright (c) 2008 Android Open Source Project
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from __future__ import with_statement

import wx
import sys
import thread
import copy
import wx.html
import operator

from wx.lib.mixins.listctrl import CheckListCtrlMixin, ListCtrlAutoWidthMixin
from eas import *
from JetFile import *
from JetUtils import *
from JetCtrls import *
from JetDialogs import *
from JetSegGraph import SegmentGraph, Marker
from JetAudition import *
from JetStatusEvent import *

import img_favicon
import img_New

provider = wx.SimpleHelpProvider()
wx.HelpProvider_Set(provider)


class JetCreator(wx.Frame):
    """ Main window of JetCreator utility """
    def __init__(self, parent, id, jetConfigFile, importFlag=False):
        wx.Frame.__init__(self, parent, id, size=(1050, 720), style=wx.DEFAULT_FRAME_STYLE | wx.MINIMIZE_BOX | wx.MAXIMIZE_BOX)

        self.myicon = img_favicon.getIcon() 
        self.SetIcon(self.myicon) 
        self.UndoStack = []
        self.RedoStack = []
        self.queueSegs = []
        self.clipBoard = None
        self.jet = None
        self.playerLock = threading.RLock()
        self.SetKeepPlayingFlag(True)
        self.currentSegmentName = None
        self.currentSegmentIndex = None
        self.currentEventName = None
        self.currentEventIndex = None
        self.currentCtrl = ""
        self.currentJetConfigFile = jetConfigFile
        self.paused = False
        self.eventlistSort = (0, 1)
        self.seglistSort = (0, 1)
        if self.currentJetConfigFile == "":
            FileKillClean(JetDefs.UNTITLED_FILE)
            self.currentJetConfigFile = JetDefs.UNTITLED_FILE
        
        self.jet_file = JetFile(self.currentJetConfigFile, "")
        
        if not ValidateConfig(self.jet_file):
            FileKillClean(JetDefs.UNTITLED_FILE)
            self.currentJetConfigFile = JetDefs.UNTITLED_FILE
            self.jet_file = JetFile(self.currentJetConfigFile, "")

        if self.currentJetConfigFile == JetDefs.UNTITLED_FILE:
            self.LoadDefaultProperties()
            
        self.initLayout()
        self.initStatusBar()
        self.createMenuBar()
        self.createToolbar()
        self.SetCurrentFile(self.currentJetConfigFile)
        self.initHelp()
        
        self.graph.ClickCallbackFct = self.GraphTriggerClip
        
        EVT_JET_STATUS(self, self.OnJetStatusUpdate)
        
        wx.EVT_CLOSE(self, self.OnClose)

        self.Centre()
        self.Show(True)
        
        if importFlag:
            self.OnJetImportArchive(None)
     
        self.eventList.OnSortOrderChangedAlert = self.OnEventSortOrderChanged
        self.segList.OnSortOrderChangedAlert = self.OnSegSortOrderChanged
        
    def initLayout(self):
        """ Initializes the screen layout """
        panel = wx.Panel(self, -1)

        hboxMain = wx.BoxSizer(wx.HORIZONTAL)

        leftPanel = wx.Panel(panel, -1)
        leftTopPanel = wx.Panel(leftPanel, -1)
        leftBotPanel = wx.Panel(leftPanel, -1)
        rightPanel = wx.Panel(panel, -1)
        
        self.segList = JetCheckListCtrl(rightPanel)
        for title, width, fld in JetDefs.SEGMENT_GRID:
            self.segList.AddCol(title, width)

        self.eventList = JetListCtrl(rightPanel)
        for title, width, fld in JetDefs.CLIPS_GRID:
            self.eventList.AddCol(title, width)

        self.eventList.Bind(wx.EVT_LIST_ITEM_SELECTED, self.OnEventListClick)
        self.segList.Bind(wx.EVT_LIST_ITEM_SELECTED, self.OnSegListClick)
        self.segList.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.OnSegmentUpdate)
        self.eventList.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.OnEventUpdate)
        
        self.segList.BindCheckBox(self.OnSegmentChecked)
        
        BUT_SIZE = (95, 25)
        self.btnAddSeg = wx.Button(leftTopPanel, -1, JetDefs.BUT_ADD, size=BUT_SIZE)
        self.btnRevSeg = wx.Button(leftTopPanel, -1, JetDefs.BUT_REVISE, size=BUT_SIZE)
        self.btnDelSeg = wx.Button(leftTopPanel, -1, JetDefs.BUT_DELETE, size=BUT_SIZE)
        self.btnMoveSeg = wx.Button(leftTopPanel, -1, JetDefs.BUT_MOVE, size=BUT_SIZE)

        self.btnQueueAll = wx.Button(leftTopPanel, -1, JetDefs.BUT_QUEUEALL, size=BUT_SIZE)
        self.btnDequeueAll = wx.Button(leftTopPanel, -1, JetDefs.BUT_DEQUEUEALL, size=BUT_SIZE)
        self.btnPlay = wx.Button(leftTopPanel, -1, JetDefs.BUT_PLAY, size=BUT_SIZE)
        self.btnPause = wx.Button(leftTopPanel, -1, JetDefs.BUT_PAUSE, size=BUT_SIZE)
        self.btnAudition = wx.Button(leftTopPanel, -1, JetDefs.BUT_AUDITION, size=BUT_SIZE)
        
        self.btnAddEvt = wx.Button(leftBotPanel, -1, JetDefs.BUT_ADD, size=BUT_SIZE)
        self.btnRevEvt = wx.Button(leftBotPanel, -1, JetDefs.BUT_REVISE, size=BUT_SIZE)
        self.btnDelEvt = wx.Button(leftBotPanel, -1, JetDefs.BUT_DELETE, size=BUT_SIZE)
        self.btnMoveEvents = wx.Button(leftBotPanel, -1, JetDefs.BUT_MOVE, size=BUT_SIZE)
        
        self.Bind(wx.EVT_BUTTON, self.OnSegmentAdd, id=self.btnAddSeg.GetId())
        self.Bind(wx.EVT_BUTTON, self.OnSegmentUpdate, id=self.btnRevSeg.GetId())
        self.Bind(wx.EVT_BUTTON, self.OnSegmentDelete, id=self.btnDelSeg.GetId())
        self.Bind(wx.EVT_BUTTON, self.OnSegmentsMove, id=self.btnMoveSeg.GetId())

        self.Bind(wx.EVT_BUTTON, self.OnSelectAll, id=self.btnQueueAll.GetId())
        self.Bind(wx.EVT_BUTTON, self.OnDeselectAll, id=self.btnDequeueAll.GetId())
        self.Bind(wx.EVT_BUTTON, self.OnPlay, id=self.btnPlay.GetId())
        self.Bind(wx.EVT_BUTTON, self.OnPause, id=self.btnPause.GetId())
        self.Bind(wx.EVT_BUTTON, self.OnAudition, id=self.btnAudition.GetId())
        
        self.Bind(wx.EVT_BUTTON, self.OnEventAdd, id=self.btnAddEvt.GetId())
        self.Bind(wx.EVT_BUTTON, self.OnEventUpdate, id=self.btnRevEvt.GetId())
        self.Bind(wx.EVT_BUTTON, self.OnEventDelete, id=self.btnDelEvt.GetId())
        self.Bind(wx.EVT_BUTTON, self.OnEventsMove, id=self.btnMoveEvents.GetId())
        
        BORDER = 5
        BUT_SPACE = 3
        vBoxLeftTop = wx.BoxSizer(wx.VERTICAL)
        vBoxLeftBot = wx.BoxSizer(wx.VERTICAL)

        vBoxLeftTop.Add(self.btnAddSeg, 0, wx.TOP, BORDER)
        vBoxLeftTop.Add(self.btnRevSeg, 0, wx.TOP, BUT_SPACE)
        vBoxLeftTop.Add(self.btnDelSeg, 0, wx.TOP, BUT_SPACE)
        vBoxLeftTop.Add(self.btnMoveSeg, 0, wx.TOP, BUT_SPACE)
        vBoxLeftTop.Add((-1, 12))
        vBoxLeftTop.Add(self.btnQueueAll, 0, wx.TOP, BUT_SPACE)
        vBoxLeftTop.Add(self.btnDequeueAll, 0, wx.TOP, BUT_SPACE)
        vBoxLeftTop.Add(self.btnPlay, 0, wx.TOP, BUT_SPACE)
        vBoxLeftTop.Add(self.btnPause, 0, wx.TOP, BUT_SPACE)
        vBoxLeftTop.Add(self.btnAudition, 0, wx.TOP, BUT_SPACE)

        vBoxLeftBot.Add(self.btnAddEvt, 0)
        vBoxLeftBot.Add(self.btnRevEvt, 0, wx.TOP, BUT_SPACE)
        vBoxLeftBot.Add(self.btnDelEvt, 0, wx.TOP, BUT_SPACE)
        vBoxLeftBot.Add(self.btnMoveEvents, 0, wx.TOP, BUT_SPACE)
        
        leftTopPanel.SetSizer(vBoxLeftTop)
        leftBotPanel.SetSizer(vBoxLeftBot)

        vboxLeft = wx.BoxSizer(wx.VERTICAL)
        vboxLeft.Add(leftTopPanel, 1, wx.EXPAND)
        vboxLeft.Add(leftBotPanel, 1, wx.EXPAND)
        vboxLeft.Add((-1, 25))
        
        leftPanel.SetSizer(vboxLeft)

        self.log = wx.TextCtrl(rightPanel, -1)
        self.graph = SegmentGraph(rightPanel, size=(-1, 50))
        
        vboxRight = wx.BoxSizer(wx.VERTICAL)
        vboxRight.Add(self.segList, 4, wx.EXPAND | wx.TOP, BORDER)
        vboxRight.Add((-1, 10))
        vboxRight.Add(self.eventList, 3, wx.EXPAND | wx.TOP, BORDER)
        vboxRight.Add((-1, 10))
        vboxRight.Add(self.log, 0, wx.EXPAND)
        vboxRight.Add((-1, 5))
        vboxRight.Add(self.graph, 1, wx.EXPAND)
        vboxRight.Add((-1, 10))

        rightPanel.SetSizer(vboxRight)

        hboxMain.Add(leftPanel, 0, wx.EXPAND | wx.RIGHT | wx.LEFT, BORDER)
        hboxMain.Add(rightPanel, 1, wx.EXPAND)
        hboxMain.Add((BORDER, -1))

        panel.SetSizer(hboxMain)

        pnlGraph = wx.Panel(leftBotPanel, -1)
        graphSizer1 = wx.BoxSizer(wx.VERTICAL)
        pnlGraph.SetSizer(graphSizer1)
        
        graphBox = wx.StaticBox(pnlGraph, wx.ID_ANY, label='Graph')
        graphSizer2 = wx.StaticBoxSizer(graphBox, wx.VERTICAL)

        self.chkGraphLabels = wx.CheckBox(pnlGraph, -1, JetDefs.GRAPH_LBLS)
        self.chkGraphClips = wx.CheckBox(pnlGraph, -1, JetDefs.GRAPH_TRIGGER)
        self.chkGraphAppEvts = wx.CheckBox(pnlGraph, -1, JetDefs.GRAPH_APP)
        
        graphSizer2.Add(self.chkGraphLabels, 0, wx.TOP, BUT_SPACE)
        graphSizer2.Add(self.chkGraphClips, 0, wx.TOP, BUT_SPACE)
        graphSizer2.Add(self.chkGraphAppEvts, 0, wx.TOP | wx.BOTTOM, BUT_SPACE)
        graphSizer1.Add((-1, 10))
        graphSizer1.Add(graphSizer2)

        vBoxLeftBot.Add(pnlGraph, 0, wx.TOP, BUT_SPACE)

        self.Bind(wx.EVT_CHECKBOX, self.OnSetGraphOptions, id=self.chkGraphLabels.GetId())
        self.Bind(wx.EVT_CHECKBOX, self.OnSetGraphOptions, id=self.chkGraphClips.GetId())
        self.Bind(wx.EVT_CHECKBOX, self.OnSetGraphOptions, id=self.chkGraphAppEvts.GetId())
        

    def initHelp(self):
        """ Initializes the help text for screen elements """
        self.SetHelpText(GetJetHelpText(JetDefs.MAIN_DLG_CTRLS, ''))
        self.segList.SetHelpText(GetJetHelpText(JetDefs.MAIN_DLG_CTRLS, JetDefs.MAIN_SEGLIST))
        self.eventList.SetHelpText(GetJetHelpText(JetDefs.MAIN_DLG_CTRLS, JetDefs.MAIN_EVENTLIST))
        self.graph.SetHelpText(GetJetHelpText(JetDefs.AUDITION_CTRLS, JetDefs.AUDITION_GRAPH))

    def initStatusBar(self):
        """ Initializes the status bar """
        self.statusbar = self.CreateStatusBar()

    def OnSelectAll(self, event):
        """ Called from select all button """
        num = self.segList.GetItemCount()
        for i in range(num-1, -1, -1):
            self.segList.CheckItem(i)

    def OnDeselectAll(self, event):
        """ Called from deselect all button """
        num = self.segList.GetItemCount()
        for i in range(num-1, -1, -1):
            self.segList.CheckItem(i, False)

    def LoadSegList(self):
        """ Loads up the list of segments """
        self.seglistSort = (IniGetValue(self.currentJetConfigFile, JetDefs.INI_SEGSORT, JetDefs.INI_SEGSORT_0, 'int', 0), IniGetValue(self.currentJetConfigFile, JetDefs.INI_SEGSORT, JetDefs.INI_SEGSORT_1, 'int', 1))
        segments = self.jet_file.GetSegments()
        if self.seglistSort[0] == 0:
            self.SegSort(segments, "segname")
        elif self.seglistSort[0] == 1:
            self.SegSort(segments, "filename")
        elif self.seglistSort[0] == 2:
            self.SegSort(segments, "dlsfile")
        elif self.seglistSort[0] == 3:
            self.SegSort(segments, "start")
        elif self.seglistSort[0] == 4:
            self.SegSort(segments, "end")
        elif self.seglistSort[0] == 5:
            self.SegSort(segments, "quantize")
        elif self.seglistSort[0] == 6:
            self.SegSort(segments, "transpose")
        elif self.seglistSort[0] == 7:
            self.SegSort(segments, "repeat")
        elif self.seglistSort[0] == 8:
            self.SegSort(segments, "mute_flags")
        listDataMap = []
        self.currentSegmentIndex = None
        self.currentSegmentName = None
        self.segList.DeleteAllItems()
        self.eventList.DeleteAllItems()
        self.menuItems[JetDefs.MNU_UPDATE_SEG].Enable(False)
        self.menuItems[JetDefs.MNU_DELETE_SEG].Enable(False)
        self.menuItems[JetDefs.MNU_MOVE_SEG].Enable(False)
        
        self.menuItems[JetDefs.MNU_ADD_EVENT].Enable(False)
        self.menuItems[JetDefs.MNU_MOVE_EVENT].Enable(False)
        self.menuItems[JetDefs.MNU_UPDATE_EVENT].Enable(False)
        self.menuItems[JetDefs.MNU_DELETE_EVENT].Enable(False)
        self.menuItems[JetDefs.MNU_MOVE_EVENT].Enable(False)
        for segment in self.jet_file.GetSegments():
            index = self.segList.InsertStringItem(sys.maxint, StrNoneChk(segment.segname))
            self.segList.SetStringItem(index, 1, FileJustName(StrNoneChk(segment.filename)))
            self.segList.SetStringItem(index, 2, FileJustName(StrNoneChk(segment.dlsfile)))
            self.segList.SetStringItem(index, 3, mbtFct(segment.start, 1))
            self.segList.SetStringItem(index, 4, mbtFct(segment.end, 1))
            self.segList.SetStringItem(index, 5, StrNoneChk(segment.quantize))
            self.segList.SetStringItem(index, 6, StrNoneChk(segment.transpose))
            self.segList.SetStringItem(index, 7, StrNoneChk(segment.repeat))
            self.segList.SetStringItem(index, 8, StrNoneChk(segment.mute_flags))
            
            self.segList.SetItemData(index, index)
            listDataMap.append((getColumnText(self.segList, index, 0).upper(), getColumnText(self.segList, index, 1).upper(), getColumnText(self.segList, index, 2).upper(), MbtVal(getColumnText(self.segList, index, 3)), MbtVal(getColumnText(self.segList, index, 4)), int(getColumnText(self.segList, index, 5)), int(getColumnText(self.segList, index, 6)), int(getColumnText(self.segList, index, 7)), int(getColumnText(self.segList, index, 8))))
            
        self.segList.itemDataMap = listDataMap
        self.segList.InitSorting(9)
        self.graph.ClearGraph()
            
    def LoadEventsForSeg(self, segName):
        """ Loads up the events associated with a segment """
        self.currentEventIndex = None
        self.eventList.DeleteAllItems()
        self.menuItems[JetDefs.MNU_UPDATE_EVENT].Enable(False)
        self.menuItems[JetDefs.MNU_DELETE_EVENT].Enable(False)
        self.menuItems[JetDefs.MNU_MOVE_EVENT].Enable(False)
        self.eventlistSort = (IniGetValue(self.currentJetConfigFile, JetDefs.INI_EVENTSORT, JetDefs.INI_EVENTSORT_0, 'int', 0), IniGetValue(self.currentJetConfigFile, JetDefs.INI_EVENTSORT, JetDefs.INI_EVENTSORT_1, 'int', 1))
        segment = self.jet_file.GetSegment(self.currentSegmentName)
        if segment is not None:
            if self.eventlistSort[0] == 0:
                self.EventSort(segment.jetevents, "event_name")
            elif self.eventlistSort[0] == 1:
                self.EventSort(segment.jetevents, "event_type")
            elif self.eventlistSort[0] == 2:
                self.EventSort(segment.jetevents, "event_start")
            elif self.eventlistSort[0] == 3:
                self.EventSort(segment.jetevents, "event_end")
            elif self.eventlistSort[0] == 4:
                self.EventSort(segment.jetevents, "track_num")
            elif self.eventlistSort[0] == 5:
                self.EventSort(segment.jetevents, "channel_num")
            elif self.eventlistSort[0] == 6:
                self.EventSort(segment.jetevents, "event_id")
            listDataMap = []
            for jet_event in self.jet_file.GetEvents(segName):
                index = self.eventList.InsertStringItem(sys.maxint, StrNoneChk(jet_event.event_name))
                self.eventList.SetStringItem(index, 1, StrNoneChk(jet_event.event_type))
                self.eventList.SetStringItem(index, 2, mbtFct(jet_event.event_start, 1))
                self.eventList.SetStringItem(index, 3, mbtFct(jet_event.event_end, 1))
                self.eventList.SetStringItem(index, 4, StrNoneChk(jet_event.track_num))
                self.eventList.SetStringItem(index, 5, StrNoneChk(jet_event.channel_num + 1))
                self.eventList.SetStringItem(index, 6, StrNoneChk(jet_event.event_id))
                
                self.eventList.SetItemData(index, index)
                listDataMap.append((getColumnText(self.eventList, index, 0).upper(), getColumnText(self.eventList, index, 1).upper(), MbtVal(getColumnText(self.eventList, index, 2)), MbtVal(getColumnText(self.eventList, index, 3)), int(getColumnText(self.eventList, index, 4)), int(getColumnText(self.eventList, index, 5)), int(getColumnText(self.eventList, index, 6))))
                    
            self.eventList.itemDataMap = listDataMap
            self.eventList.InitSorting(7)

    def OnEventListClick(self, event):
        """ Sets the current event variable when selecting from the list """
        self.currentCtrl = "eventList"
        self.currentEventIndex = event.m_itemIndex
        self.currentEventName = getColumnText(self.eventList, event.m_itemIndex, 0)       
        self.menuItems[JetDefs.MNU_UPDATE_EVENT].Enable(True)
        self.menuItems[JetDefs.MNU_DELETE_EVENT].Enable(True)
        self.menuItems[JetDefs.MNU_MOVE_EVENT].Enable(True)
               
    def OnSegmentChecked(self, index, checked):
        """ Selects the segment when checkbox clicked """
        ClearRowSelections(self.segList)
        SetRowSelection(self.segList, index, True)
        
    def SelectSegment(self, segName):
        """ Selects a segment by segment name """
        itm = self.segList.FindItem(-1, segName)
        self.segList.EnsureVisible(itm)
        ClearRowSelections(self.segList)
        SetRowSelection(self.segList, itm, True)
        
    def SelectEvent(self, eventName):
        """ Selects an event by event name """
        itm = self.eventList.FindItem(-1, eventName)
        self.eventList.EnsureVisible(itm)
        ClearRowSelections(self.eventList)
        SetRowSelection(self.eventList, itm, True)
        
    def OnSegListClick(self, event):
        """ Loads up a segment when the list is clicked """
        self.currentCtrl = "segList"
        self.currentSegmentIndex = event.m_itemIndex
        self.currentSegmentName = getColumnText(self.segList, event.m_itemIndex, 0)
        self.LoadEventsForSeg(getColumnText(self.segList, event.m_itemIndex, 0))
        self.menuItems[JetDefs.MNU_UPDATE_SEG].Enable(True)
        self.menuItems[JetDefs.MNU_DELETE_SEG].Enable(True)
        self.menuItems[JetDefs.MNU_MOVE_SEG].Enable(True)
        self.menuItems[JetDefs.MNU_ADD_EVENT].Enable(True)
        info = self.graph.LoadSegment(self.jet_file.GetSegment(self.currentSegmentName), showLabels=IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHLABELS, JetDefs.F_GRAPHLABELS, 'bool', 'True'), showClips=IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHCLIPS, JetDefs.F_GRAPHCLIPS, 'bool', 'True'), showAppEvts=IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHAPPEVTS, JetDefs.F_GRAPHAPPEVTS, 'bool', 'True'))
        if info == None:
            self.log.SetValue("")
        else:
            iLength = info.iLengthInMs
            if iLength > 0:
                self.log.SetValue("%s      %.2f Seconds" % (self.currentSegmentName, iLength / 1000.00))
            else:
                self.log.SetValue("%s" % (self.currentSegmentName))    
    
    def OnSegmentAdd(self, event):
        """ Calls the dialog box for adding a segment """
        saveState = JetState(self.jet_file, self.currentSegmentIndex, self.currentEventIndex)
        
        dlg = SegEdit(JetDefs.MAIN_ADDSEGTITLE, self.currentJetConfigFile)
       
        for filename in self.jet_file.GetMidiFiles():
            dlg.je.ctrls[JetDefs.F_MIDIFILE].Append(filename)
        for library in self.jet_file.GetLibraries():
            dlg.je.ctrls[JetDefs.F_DLSFILE].Append(library)
        
        result = dlg.ShowModal()
        if result == wx.ID_OK:
            if len(dlg.lstReplicate) > 0:
                if dlg.chkReplaceMatching:
                    self.jet_file.DeleteSegmentsMatchingPrefix(dlg.replicatePrefix) 
                
                for replicate in dlg.lstReplicate:
                    self.jet_file.AddSegment(replicate[0], dlg.GetValue(JetDefs.F_MIDIFILE), 
                                             mbtFct(replicate[1],-1), mbtFct(replicate[2],-1), 
                                             JetDefs.MBT_ZEROSTR, 
                                             SegmentOutputFile(dlg.GetValue(JetDefs.F_SEGNAME), self.currentJetConfigFile), 
                                             dlg.GetValue(JetDefs.F_QUANTIZE),
                                             [], dlg.GetValue(JetDefs.F_DLSFILE),
                                             None,
                                             dlg.GetValue(JetDefs.F_TRANSPOSE),
                                             dlg.GetValue(JetDefs.F_REPEAT),
                                             dlg.GetValue(JetDefs.F_MUTEFLAGS))

                self.LoadSegList()
                self.SelectSegment(dlg.lstReplicate[0][0])
            else:
                self.jet_file.AddSegment(dlg.GetValue(JetDefs.F_SEGNAME), dlg.GetValue(JetDefs.F_MIDIFILE), 
                                     dlg.GetValue(JetDefs.F_START), dlg.GetValue(JetDefs.F_END), 
                                     JetDefs.MBT_ZEROSTR, 
                                     SegmentOutputFile(dlg.GetValue(JetDefs.F_SEGNAME), self.currentJetConfigFile), 
                                     dlg.GetValue(JetDefs.F_QUANTIZE),
                                     [], dlg.GetValue(JetDefs.F_DLSFILE),
                                     None,
                                     dlg.GetValue(JetDefs.F_TRANSPOSE),
                                     dlg.GetValue(JetDefs.F_REPEAT),
                                     dlg.GetValue(JetDefs.F_MUTEFLAGS))
                self.LoadSegList()
                self.SelectSegment(dlg.GetValue(JetDefs.F_SEGNAME))
            self.UndoAdd(saveState)
        dlg.Destroy()
        
    def OnSegmentUpdate(self, event):
        """ Calls the dialog box for updating a segment """
        if self.currentSegmentName is None:
            return
        
        segment = self.jet_file.GetSegment(self.currentSegmentName)
        if segment == None:
            return
        
        saveState = JetState(self.jet_file, self.currentSegmentIndex, self.currentEventIndex)
        
        dlg = SegEdit(JetDefs.MAIN_REVSEGTITLE, self.currentJetConfigFile)
       
        for filename in self.jet_file.GetMidiFiles():
            dlg.je.ctrls[JetDefs.F_MIDIFILE].Append(filename)
        for library in self.jet_file.GetLibraries():
            dlg.je.ctrls[JetDefs.F_DLSFILE].Append(library)
                    
        dlg.SetValue(JetDefs.F_SEGNAME, segment.segname)
        dlg.SetValue(JetDefs.F_MUTEFLAGS, segment.mute_flags)
        dlg.SetValue(JetDefs.F_MIDIFILE, segment.filename)
        dlg.SetValue(JetDefs.F_DLSFILE, segment.dlsfile)
        dlg.SetValue(JetDefs.F_START, segment.start)
        dlg.SetValue(JetDefs.F_END, segment.end)
        dlg.SetValue(JetDefs.F_QUANTIZE, segment.quantize)
        dlg.SetValue(JetDefs.F_TRANSPOSE, segment.transpose)
        dlg.SetValue(JetDefs.F_REPEAT, segment.repeat)
        dlg.jetevents = segment.jetevents
        
        result = dlg.ShowModal()
        if result == wx.ID_OK:
            self.jet_file.UpdateSegment(self.currentSegmentName, dlg.GetValue(JetDefs.F_SEGNAME), 
                                     dlg.GetValue(JetDefs.F_MIDIFILE), 
                                     dlg.GetValue(JetDefs.F_START), dlg.GetValue(JetDefs.F_END), 
                                     JetDefs.MBT_ZEROSTR, #dlg.GetValue(JetDefs.F_LENGTH), 
                                     SegmentOutputFile(dlg.GetValue(JetDefs.F_SEGNAME), self.currentJetConfigFile), 
                                     dlg.GetValue(JetDefs.F_QUANTIZE),
                                     [], dlg.GetValue(JetDefs.F_DLSFILE),
                                     None,
                                     dlg.GetValue(JetDefs.F_TRANSPOSE),
                                     dlg.GetValue(JetDefs.F_REPEAT),
                                     dlg.GetValue(JetDefs.F_MUTEFLAGS))
            
            if len(dlg.lstReplicate) > 0:
                if dlg.chkReplaceMatching:
                    self.jet_file.DeleteSegmentsMatchingPrefix(dlg.replicatePrefix) 
                
                for replicate in dlg.lstReplicate:
                    self.jet_file.AddSegment(replicate[0], dlg.GetValue(JetDefs.F_MIDIFILE), 
                                             mbtFct(replicate[1],-1), mbtFct(replicate[2],-1), 
                                             JetDefs.MBT_ZEROSTR, 
                                             SegmentOutputFile(dlg.GetValue(JetDefs.F_SEGNAME), self.currentJetConfigFile), 
                                             dlg.GetValue(JetDefs.F_QUANTIZE),
                                             [], dlg.GetValue(JetDefs.F_DLSFILE),
                                             None,
                                             dlg.GetValue(JetDefs.F_TRANSPOSE),
                                             dlg.GetValue(JetDefs.F_REPEAT),
                                             dlg.GetValue(JetDefs.F_MUTEFLAGS))
    
                self.LoadSegList()
                self.SelectSegment(dlg.lstReplicate[0][0])
            else:          
                self.LoadSegList()
                self.SelectSegment(dlg.GetValue(JetDefs.F_SEGNAME))
            self.UndoAdd(saveState)
        dlg.Destroy()
        
    def OnSegmentDelete(self, event):
        """ Confirms the deletion segment(s) by user action """
        if self.currentSegmentName is None:
            return
        
        segment = self.jet_file.GetSegment(self.currentSegmentName)
        if segment == None:
            return
        
        count = 0
        deleteMsg = ''
        item = self.segList.GetFirstSelected()
        while item != -1:
            if count == 0:
                deleteMsg = getColumnText(self.segList,item,0)
            else:
                if count == 40:
                    deleteMsg = deleteMsg + "\n" + "....more"
                elif count < 40:
                    deleteMsg = deleteMsg + "\n" + getColumnText(self.segList,item,0)
            count = count + 1
            item = self.segList.GetNextSelected(item)
        
        if YesNo(JetDefs.MAIN_CONFIRM, deleteMsg + JetDefs.MAIN_CONFIRM_SEG_DLT, False):
            item = self.segList.GetFirstSelected()
            while item != -1:
                segName = getColumnText(self.segList,item,0)
                self.SegmentDelete(segName)
                item = self.segList.GetNextSelected(item)
            
            self.graph.ClearGraph()
            self.LoadSegList()
         
    def SegmentDelete(self, segName):
        """ Deletes a segment """
        saveState = JetState(self.jet_file, self.currentSegmentIndex, self.currentEventIndex)
        self.jet_file.DeleteSegment(segName)
        self.UndoAdd(saveState)
    
    def OnSegmentsMove(self, event):
        """ Move segment(s) """
        if self.currentSegmentName is None:
            return
        
        lstMoveItems = []
        count = 0
        item = self.segList.GetFirstSelected()
        while item != -1:
            max = GetMidiInfo(self.jet_file.GetSegment(getColumnText(self.segList,item,0)).filename).endMbtStr
            lstMoveItems.append((getColumnText(self.segList,item,0), mbtFct(getColumnText(self.segList,item,3),-1), mbtFct(getColumnText(self.segList,item,4),-1), max))
            count = count + 1
            item = self.segList.GetNextSelected(item)

        if count == 0:
            InfoMsg("Move", "Select one or more items to move.")
            return
        
        dlg = JetMove("Move Segments")
        dlg.lstMoveItems = lstMoveItems
        result = dlg.ShowModal()
        if result == wx.ID_OK:
            if len(dlg.lstMoveMbt) > 0:
                saveState = JetState(self.jet_file, self.currentSegmentIndex, self.currentEventIndex)
                
                for moveitem in dlg.lstMoveMbt:
                    self.jet_file.MoveSegment(moveitem[0], moveitem[1], moveitem[2])
                
                self.LoadSegList()
                self.UndoAdd(saveState)
                
        dlg.Destroy()
        
    def UndoAdd(self, saveState):
        """ Adds the current state to the undo stack """
        self.UndoStack.append(saveState)
        self.menuItems[JetDefs.MNU_UNDO].Enable(True)
       
    def RedoAdd(self, saveState):
        """ Adds the current state the the redo stack """
        self.RedoStack.append(saveState)
        self.menuItems[JetDefs.MNU_REDO].Enable(True)
       
    def OnRedo(self, event):
        """ Redo if there's one in the stack """
        if len(self.RedoStack) > 0:
            self.UndoAdd(JetState(self.jet_file, self.currentSegmentIndex, self.currentEventIndex))
            state = self.RedoStack.pop()
            self.jet_file = copy.deepcopy(state.jet_file)
            self.LoadSegList()
            self.currentSegmentIndex = state.currentSegmentIndex
            if self.currentSegmentIndex != None:
                SetRowSelection(self.segList, self.currentSegmentIndex, True)
            if len(self.RedoStack) == 0:
                self.menuItems[JetDefs.MNU_REDO].Enable(False)
                
    def OnUndo(self, event):
        """ Undo if there's one in the stack """
        if len(self.UndoStack) > 0:
            self.RedoAdd(JetState(self.jet_file, self.currentSegmentIndex, self.currentEventIndex))
            state = self.UndoStack.pop()
            self.jet_file = copy.deepcopy(state.jet_file)
            self.LoadSegList()
            self.currentSegmentIndex = state.currentSegmentIndex
            if self.currentSegmentIndex != None:
                SetRowSelection(self.segList, self.currentSegmentIndex, True)
            if len(self.UndoStack) == 0:
                self.menuItems[JetDefs.MNU_UNDO].Enable(False)

    def OnEventAdd(self, event):
        """ Calls a dialog box to add an event to the current segment """
        if self.currentSegmentName is None:
            return
        
        segment = self.jet_file.GetSegment(self.currentSegmentName)
        if segment == None:
            return
        
        saveState = JetState(self.jet_file, self.currentSegmentIndex, self.currentEventIndex)
        
        dlg = EventEdit(JetDefs.MAIN_ADDEVENTTITLE, self.currentJetConfigFile)
        editSegment = copy.deepcopy(segment)
        dlg.SetSegment(editSegment)
        dlg.SetEventId()
                      
        result = dlg.ShowModal()
        if result == wx.ID_OK:
            if dlg.GetValue(JetDefs.F_ETYPE) == JetDefs.E_EOS:
                #check for an existing EOS event
                events = self.jet_file.GetEvents(self.currentSegmentName)
                for evt in events:
                    if evt.event_type == JetDefs.E_EOS:
                        self.jet_file.DeleteEvent(self.currentSegmentName, evt.event_name)
                dlg.SetValue(JetDefs.F_ESTART, dlg.GetValue(JetDefs.F_EEND))
            
            if len(dlg.lstReplicate) > 0:
                if dlg.chkReplaceMatching:
                    self.jet_file.DeleteEventsMatchingPrefix(self.currentSegmentName, dlg.replicatePrefix) 
                
                for replicate in dlg.lstReplicate:
                    self.jet_file.AddEvent(self.currentSegmentName, replicate[0],
                                         dlg.GetValue(JetDefs.F_ETYPE), 
                                         dlg.GetValue(JetDefs.F_EEVENTID), 
                                         dlg.GetValue(JetDefs.F_ETRACK), 
                                         dlg.GetValue(JetDefs.F_ECHANNEL), 
                                         mbtFct(replicate[1],-1),
                                         mbtFct(replicate[2],-1))
                self.SelectSegment(self.currentSegmentName)
                self.SelectEvent(dlg.lstReplicate[0][0])
            else:
                self.jet_file.AddEvent(self.currentSegmentName, dlg.GetValue(JetDefs.F_ENAME),
                                     dlg.GetValue(JetDefs.F_ETYPE), 
                                     dlg.GetValue(JetDefs.F_EEVENTID), 
                                     dlg.GetValue(JetDefs.F_ETRACK), 
                                     dlg.GetValue(JetDefs.F_ECHANNEL), 
                                     dlg.GetValue(JetDefs.F_ESTART),
                                     dlg.GetValue(JetDefs.F_EEND))
            
                self.SelectSegment(self.currentSegmentName)
                self.SelectEvent(dlg.GetValue(JetDefs.F_ENAME))
                
            self.UndoAdd(saveState)
        dlg.Destroy()
        
    def OnEventUpdate(self, event):
        """ Calls the dialog box to update the current event """
        if self.currentSegmentName is None:
            return

        if self.currentEventName  is None:
            return
        
        segment = self.jet_file.GetSegment(self.currentSegmentName)
        if segment == None:
            return
        
        curEvent = copy.deepcopy(self.jet_file.GetEvent(self.currentSegmentName, self.currentEventName))
        if curEvent == None:
            return
        
        saveState = JetState(self.jet_file, self.currentSegmentIndex, self.currentEventIndex)
        
        #only want the event we are editing to show up in graph
        editSegment = copy.deepcopy(segment)
        editSegment.jetevents = []
        editSegment.jetevents.append(curEvent)
        
        dlg = EventEdit(JetDefs.MAIN_REVEVENTTITLE, self.currentJetConfigFile)
        dlg.SetSegment(editSegment)
        dlg.SetValue(JetDefs.F_ENAME, curEvent.event_name)
        dlg.SetValue(JetDefs.F_ETYPE, curEvent.event_type)
        dlg.SetValue(JetDefs.F_ESTART, curEvent.event_start)
        dlg.SetValue(JetDefs.F_EEND, curEvent.event_end)
        dlg.SetValue(JetDefs.F_ETRACK, curEvent.track_num)
        dlg.SetValue(JetDefs.F_ECHANNEL, curEvent.channel_num)
        dlg.SetValue(JetDefs.F_EEVENTID, curEvent.event_id)
        dlg.OnEventSelect()   
        
        result = dlg.ShowModal()
        if result == wx.ID_OK:
            if dlg.GetValue(JetDefs.F_ETYPE) == JetDefs.E_EOS:
                dlg.SetValue(JetDefs.F_ESTART, dlg.GetValue(JetDefs.F_EEND))
                
            self.jet_file.UpdateEvent(self.currentSegmentName, 
                                      self.currentEventName, 
                                     dlg.GetValue(JetDefs.F_ENAME),
                                     dlg.GetValue(JetDefs.F_ETYPE), 
                                     dlg.GetValue(JetDefs.F_EEVENTID), 
                                     dlg.GetValue(JetDefs.F_ETRACK), 
                                     dlg.GetValue(JetDefs.F_ECHANNEL), 
                                     dlg.GetValue(JetDefs.F_ESTART),
                                     dlg.GetValue(JetDefs.F_EEND))

            if len(dlg.lstReplicate) > 0:
                if dlg.chkReplaceMatching:
                    self.jet_file.DeleteEventsMatchingPrefix(self.currentSegmentName, dlg.replicatePrefix) 
                
                for replicate in dlg.lstReplicate:
                    self.jet_file.AddEvent(self.currentSegmentName, replicate[0],
                                         dlg.GetValue(JetDefs.F_ETYPE), 
                                         dlg.GetValue(JetDefs.F_EEVENTID), 
                                         dlg.GetValue(JetDefs.F_ETRACK), 
                                         dlg.GetValue(JetDefs.F_ECHANNEL), 
                                         mbtFct(replicate[1],-1),
                                         mbtFct(replicate[2],-1))
                self.SelectSegment(self.currentSegmentName)
                self.SelectEvent(dlg.lstReplicate[0][0])
            else:            
                self.SelectSegment(self.currentSegmentName)
                self.SelectEvent(dlg.GetValue(JetDefs.F_ENAME))
            self.UndoAdd(saveState)
        dlg.Destroy()
        
    def OnEventDelete(self, event):
        """ Confirms the deletion of event(s) """
        if self.currentSegmentName is None:
            return
        
        if self.currentEventName  is None:
            return
        
        curEvent = self.jet_file.GetEvent(self.currentSegmentName, self.currentEventName)
        if curEvent == None:
            return

        count = 0
        deleteMsg = ''
        item = self.eventList.GetFirstSelected()
        while item != -1:
            if count == 0:
                deleteMsg = getColumnText(self.eventList,item,0)
            else:
                if count == 40:
                    deleteMsg = deleteMsg + "\n" + "....more"
                elif count < 40:
                    deleteMsg = deleteMsg + "\n" + getColumnText(self.eventList,item,0)
            count = count + 1
            item = self.eventList.GetNextSelected(item)
        

        if YesNo(JetDefs.MAIN_CONFIRM, deleteMsg + JetDefs.MAIN_CONRIRM_EVT_DLT, False):
            item = self.eventList.GetFirstSelected()
            while item != -1:
                eventName = getColumnText(self.eventList,item,0)
                self.EventDelete(eventName)
                item = self.eventList.GetNextSelected(item)

            self.SelectSegment(self.currentSegmentName)
            self.LoadEventsForSeg(self.currentSegmentName)
          
    def EventDelete(self, eventName):
        """ Deletes an event """
        saveState = JetState(self.jet_file, self.currentSegmentIndex, self.currentEventIndex)
        self.jet_file.DeleteEvent(self.currentSegmentName, eventName)
        self.UndoAdd(saveState)

    def OnEventsMove(self, event):
        """ Move event(s) """
        if self.currentSegmentName is None:
            return
        
        if self.currentEventName  is None:
            return
        
        segment = self.jet_file.GetSegment(self.currentSegmentName)
        if segment == None:
            return
        
        curEvent = self.jet_file.GetEvent(self.currentSegmentName, self.currentEventName)
        if curEvent == None:
            return

        lstMoveItems = []
        count = 0
        item = self.eventList.GetFirstSelected()
        while item != -1:
            lstMoveItems.append((getColumnText(self.eventList,item,0), mbtFct(getColumnText(self.eventList,item,2),-1), mbtFct(getColumnText(self.eventList,item,3),-1), segment.end))
            count = count + 1
            item = self.eventList.GetNextSelected(item)

        if count == 0:
            InfoMsg("Move", "Select one or more items to move.")
            return
        
        dlg = JetMove("Move Events")
        dlg.lstMoveItems = lstMoveItems
        result = dlg.ShowModal()
        if result == wx.ID_OK:
            if len(dlg.lstMoveMbt) > 0:
                saveState = JetState(self.jet_file, self.currentSegmentIndex, self.currentEventIndex)
                
                for moveitem in dlg.lstMoveMbt:
                    self.jet_file.MoveEvent(self.currentSegmentName, moveitem[0], moveitem[1], moveitem[2])
                
                self.SelectSegment(self.currentSegmentName)
                self.LoadEventsForSeg(self.currentSegmentName)
                
                self.UndoAdd(saveState)
                
        dlg.Destroy()
        
    def OnJetOpen(self, event):
        """ Calls a dialog box to get a jet config file to open """
        dlg = JetOpen()
        result = dlg.ShowModal()
        if result == JetDefs.ID_JET_OPEN:
            self.jet_file = JetFile(dlg.fileName , "")
            if not ValidateConfig(self.jet_file):
                FileKillClean(JetDefs.UNTITLED_FILE)
                self.currentJetConfigFile = JetDefs.UNTITLED_FILE
                self.jet_file = JetFile(self.currentJetConfigFile, "")
            else:
                self.SetCurrentFile(dlg.fileName)
        elif result == JetDefs.ID_JET_NEW:
            self.jet_file = JetFile("" , "")
            self.SetCurrentFile(JetDefs.UNTITLED_FILE)
            self.LoadDefaultProperties()
        elif result == JetDefs.ID_JET_IMPORT:
            self.OnJetImportArchive(event)
        dlg.Destroy()

    def OnJetSaveAs(self, event): 
        """ Calls a dialog box to allow saving the current jet file as another name """
        defDir = IniGetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_DEFAULTDIRS, JetDefs.JTC_FILE_SPEC, 'str', str(os.getcwd()))
        dialog = wx.FileDialog(None, JetDefs.SAVE_PROMPT, defDir, "", JetDefs.JTC_FILE_SPEC, wx.SAVE | wx.OVERWRITE_PROMPT )
        if dialog.ShowModal() == wx.ID_OK:
            IniSetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_DEFAULTDIRS, JetDefs.JTC_FILE_SPEC, str(FileJustPath(dialog.GetPath())))
            self.currentJetConfigFile = FileJustRoot(dialog.GetPath()) + ".jtc"            
            self.jet_file.config.filename = FileJustRoot(self.currentJetConfigFile)  + ".jet"
            self.jet_file.SaveJetConfig(self.currentJetConfigFile)
            self.jet_file.WriteJetFileFromConfig(self.currentJetConfigFile)
            self.SetCurrentFile(self.currentJetConfigFile)
        dialog.Destroy()
    
    def OnJetSave(self, event):
        """ Saves the current jet file to disk """
        if self.currentJetConfigFile == JetDefs.UNTITLED_FILE:
            self.OnJetSaveAs(event)
        else:
            self.jet_file.SaveJetConfig(self.currentJetConfigFile)
            self.jet_file.WriteJetFileFromConfig(self.currentJetConfigFile)
        
    def OnJetNew(self, event):
        """ Initializes the state to a new jet file """
        self.jet_file = JetFile("" , "")
        self.SetCurrentFile(JetDefs.UNTITLED_FILE)
        self.LoadDefaultProperties()
                
    def SetCurrentFile(self, fileName):
        """ Sets the state for the current jet file """
        self.clipBoard = None
        self.currentJetConfigFile = fileName
        self.SetTitle(JetDefs.MAIN_TITLEPREFIX + FileJustName(fileName))
        AppendRecentJetFile(fileName)
        self.LoadSegList()
        self.graph.ClearGraph()
        self.chkGraphLabels.SetValue(IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHLABELS, JetDefs.F_GRAPHLABELS, 'bool', 'True'))
        self.chkGraphClips.SetValue(IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHCLIPS, JetDefs.F_GRAPHCLIPS, 'bool', 'True'))
        self.chkGraphAppEvts.SetValue(IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHAPPEVTS, JetDefs.F_GRAPHAPPEVTS, 'bool', 'True'))

    def createMenuBar(self):
        """ Creates a menu bar """
        self.menuItems = {}
        menuBar = wx.MenuBar()
        for eachMenuData in JetDefs.MENU_SPEC:
            menuLabel = eachMenuData[0]
            menuItems = eachMenuData[1:]
            menuBar.Append(self.createMenu(menuItems), menuLabel)
        self.SetMenuBar(menuBar)

    def createMenu(self, menuData):
        """ Creates a menu from the structure menuData in JetDefs """
        menu = wx.Menu()
        for eachLabel, eachStatus, eachHandler, eachEnable in menuData:
            if not eachLabel:
                menu.AppendSeparator()
                continue
            self.menuItems[eachLabel] = menu.Append(-1, eachLabel, eachStatus)
            self.menuItems[eachLabel].Enable(eachEnable)
            try:
                self.Bind(wx.EVT_MENU, getattr(self, eachHandler) , self.menuItems[eachLabel])
            except:
                print("def " + eachHandler + "(self, event): pass")  
        return menu

    def createToolbar(self):
        """ Creates the toolbar """
        toolbar = self.CreateToolBar()
        toolbar.SetToolBitmapSize((32,32))
        self.toolItems = {}
        for eachTool in JetDefs.TOOLBAR_SPEC:
            if eachTool[0] == '-':
                toolbar.AddSeparator()
            else:
                b = __import__(eachTool[1])
                bitMap = b.getBitmap()
                self.toolItems[eachTool[0]] = toolbar.AddLabelTool(-1, label=eachTool[0], 
                        bitmap=bitMap, 
                                     shortHelp=eachTool[0], longHelp=eachTool[2])
                self.Bind(wx.EVT_TOOL, getattr(self, eachTool[3]) , self.toolItems[eachTool[0]])
        toolbar.Realize()
        
    def OnAudition(self, event):
        """ Calls the audition window for simple preview of jet file """
        jet_file = CreateTempJetFile(self.jet_file)

        w, h = self.GetSize()
        w = w - 50
        if w < 900:
            w = 900
        h = h - 50
        if h < 650:
            h = 650
        dlg = Audition(jet_file, (w,h))
        dlg.ShowModal()   
        CleanupTempJetFile(jet_file)
        
    def SetKeepPlayingFlag(self, val):
        """ Sets a flag to communicate playing state to the play thread """
        with self.playerLock:
            self.keepPlaying = val

    def GetKeepPlayingFlag(self):   
        """ Gets the playing state flag """   
        with self.playerLock:
            return self.keepPlaying
        
    def GraphTriggerClip(self, sClipName, iEventId):
        """ Triggers a clip when they click on the graph """
        with self.playerLock:
            try:
                self.jet.TriggerClip(iEventId)
                self.log.SetValue(JetDefs.PLAY_TRIGGERCLIP_MSG % (iEventId, sClipName))
                return True
            except:
                return False       
        
    def OnHelpJet(self, event):
        """ Loads the jet help file """
        import webbrowser
        webbrowser.open(JetDefs.MAIN_HELPFILE)
        return

    def OnHelpJetGuidelines(self, event):
        """ Loads the authoring guidelines file """
        import webbrowser
        webbrowser.open(JetDefs.MAIN_HELPGUIDELINESFILE)
        return
            
    def OnAbout(self, event): 
        """ Loads the about dialog box """
        dlg = JetAbout()
        result = dlg.ShowModal()
        dlg.Destroy()

    def OnJetImportArchive(self, event):
        """ Imports a jet archive file """
        defDir = IniGetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_DEFAULTDIRS, JetDefs.ARCHIVE_FILE_SPEC, 'str', str(os.getcwd()))
        dialog = wx.FileDialog(None, JetDefs.IMPORT_ARCHIVE_PROMPT, defDir, "", JetDefs.ARCHIVE_FILE_SPEC, wx.OPEN)
        if dialog.ShowModal() == wx.ID_OK:
            IniSetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_DEFAULTDIRS, JetDefs.ARCHIVE_FILE_SPEC, str(FileJustPath(dialog.GetPath())))
            defDir = IniGetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_DEFAULTDIRS, JetDefs.ARCHIVE_FILE_SPEC + "Dir", 'str', str(os.getcwd()))
            dlg1 = wx.DirDialog(self, JetDefs.IMPORT_ARCHIVEDIR_PROMPT, style=wx.DD_DEFAULT_STYLE, defaultPath=defDir)
            if dlg1.ShowModal() == wx.ID_OK:
                IniSetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_DEFAULTDIRS, JetDefs.ARCHIVE_FILE_SPEC + "Dir", str(FileJustPath(dlg1.GetPath())))
                if YesNo(JetDefs.MAIN_IMPORTTITLE, JetDefs.MAIN_IMPORTMSG % (dialog.GetPath(),dlg1.GetPath()), False):
                    projectPath = dlg1.GetPath()
                    zipFile = dialog.GetPath()
                    z = __import__('zipfile')
                    
                    if not z.is_zipfile(zipFile):
                        wx.MessageBox(JetDefs.IMPORT_ARCHIVE_NO_JTC)
                    else:
                        zip = z.ZipFile(zipFile, 'r')
                        
                        jtcFile = ""
                        fileList = zip.namelist()
                        
                        isArchive = False
                        for myFile in fileList:
                            if myFile == 'JetArchive':
                                isArchive = True
                                break
                        if not isArchive:
                            wx.MessageBox(JetDefs.IMPORT_NOT_JET_ARCHIVE)
                        else:
                            for myFile in fileList:
                                if FileJustExt(myFile) == '.JTC':
                                    jtcFile = myFile
                                    break
                            if jtcFile == "":
                                wx.MessageBox(JetDefs.IMPORT_ARCHIVE_NO_JTC)
                            else:
                                for name in zip.namelist(): 
                                    ext = FileJustExt(name)
                                    if ext == '.MID' or ext == '.DLS'  or ext == '.JET':
                                        file(FileFixPath(projectPath + "/" + name), 'wb').write(zip.read(name))
                                    else:
                                        if len(ext) > 0 and ext != '.DS_STORE':
                                            file(FileFixPath(projectPath + "/" + name), 'w').write(zip.read(name))
                            zip.close()
                            self.currentJetConfigFile = FileFixPath(projectPath + "/") + jtcFile
                            self.jet_file = JetFile(self.currentJetConfigFile, "")
                            
                            #fix paths
                            self.jet_file.config.filename = FileJustRoot(self.currentJetConfigFile) + ".JET"
                            for index, segment in enumerate(self.jet_file.segments):
                                self.jet_file.segments[index].filename = FileFixPath(projectPath + "/" + segment.filename)
                                if segment.dlsfile > "":
                                    self.jet_file.segments[index].dlsfile = FileFixPath(projectPath + "/" + segment.dlsfile)
                                self.jet_file.segments[index].output = FileFixPath(projectPath + "/" + segment.output)                        
                            
                            for index, library in enumerate(self.jet_file.libraries):
                                self.jet_file.libraries[index] = FileFixPath(projectPath + "/" + library)
                                
                            if ValidateConfig(self.jet_file):
                                self.jet_file.SaveJetConfig(self.currentJetConfigFile)
                                self.jet_file.WriteJetFileFromConfig(self.currentJetConfigFile)
                                self.jet_file = JetFile(self.currentJetConfigFile , "")
                                self.SetCurrentFile(self.currentJetConfigFile)

            dlg1.Destroy()
        dialog.Destroy()        
       
    def OnJetExportArchive(self, event):
        """ Exports the current setup to an archive file """
        self.OnJetSave(event)
        defDir = IniGetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_DEFAULTDIRS, JetDefs.ARCHIVE_FILE_SPEC, 'str', str(os.getcwd()))
        dialog = wx.FileDialog(None, JetDefs.EXPORT_ARCHIVE_PROMPT, defDir, "", JetDefs.ARCHIVE_FILE_SPEC, wx.SAVE | wx.OVERWRITE_PROMPT )
        if dialog.ShowModal() == wx.ID_OK:
            IniSetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_DEFAULTDIRS, JetDefs.ARCHIVE_FILE_SPEC, str(FileJustPath(dialog.GetPath())))
            ExportJetArchive(FileJustRoot(dialog.GetPath()) + ".zip", self.currentJetConfigFile, self.jet_file)
        dialog.Destroy()
        
    def OnCopy(self, event):
        """ Copies the current segment or event to the clipboard """
        if self.currentCtrl == JetDefs.MAIN_SEGLIST:
            if self.currentSegmentName is None:
                return ""
            
            segment = self.jet_file.GetSegment(self.currentSegmentName)
            if segment == None:
                return ""
            self.clipBoard = JetCutCopy(self.currentCtrl, segment, self.currentSegmentName)
            return self.currentCtrl
        elif self.currentCtrl == JetDefs.MAIN_EVENTLIST:
            if self.currentSegmentName is None:
                return ""
    
            if self.currentEventName  is None:
                return ""
            
            segment = self.jet_file.GetSegment(self.currentSegmentName)
            if segment == None:
                return ""
            
            curEvent = self.jet_file.GetEvent(self.currentSegmentName, self.currentEventName)
            if curEvent == None:
                return ""
            self.clipBoard = JetCutCopy(self.currentCtrl, curEvent, self.currentSegmentName)
            return self.currentCtrl
            
    def OnCut(self, event):
        """ Cuts the current segment or event to the clipboard """
        retCopy = self.OnCopy(event)
        if retCopy == JetDefs.MAIN_SEGLIST:
            self.SegmentDelete(self.currentSegmentName) 
            self.LoadSegList()
        elif retCopy == JetDefs.MAIN_EVENTLIST:
            self.EventDelete(self.currentEventName)
            self.LoadEventsForSeg(self.currentSegmentName)
        
    def OnPaste(self, event):
        """ Pastes the current segment or event from the clipboard """
        if self.clipBoard is not None:
            if self.currentCtrl == JetDefs.MAIN_SEGLIST and self.clipBoard.objType == JetDefs.MAIN_SEGLIST:
                saveState = JetState(self.jet_file, self.currentSegmentIndex, self.currentEventIndex)
                self.jet_file.segments.append(self.clipBoard.GetObj(self.jet_file.segments))
                self.UndoAdd(saveState)
                self.LoadSegList()
            elif self.currentCtrl == JetDefs.MAIN_EVENTLIST and self.clipBoard.objType == JetDefs.MAIN_EVENTLIST and self.clipBoard.currentSegmentName == self.currentSegmentName:
                for segment in self.jet_file.segments:
                    if segment.segname == self.currentSegmentName:
                        saveState = JetState(self.jet_file, self.currentSegmentIndex, self.currentEventIndex)
                        segment.jetevents.append(self.clipBoard.GetObj(segment.jetevents))
                        self.UndoAdd(saveState)
                        self.LoadEventsForSeg(self.currentSegmentName)
                        self.graph.LoadSegment(self.jet_file.GetSegment(self.currentSegmentName), showLabels=IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHLABELS, JetDefs.F_GRAPHLABELS, 'bool', 'True'), showClips=IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHCLIPS, JetDefs.F_GRAPHCLIPS, 'bool', 'True'), showAppEvts=IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHAPPEVTS, JetDefs.F_GRAPHAPPEVTS, 'bool', 'True'))
                        
    def OnJetProperties(self, event):
        """ Opens a dialog box for jetfile properties """
        dlg = JetPropertiesDialog()
        dlg.SetValue(JetDefs.F_JETFILENAME, self.jet_file.config.filename)
        dlg.SetValue(JetDefs.F_COPYRIGHT, StrNoneChk(self.jet_file.config.copyright))
        dlg.SetValue(JetDefs.F_CHASECONTROLLERS, StrNoneChk(self.jet_file.config.chase_controllers))
        dlg.SetValue(JetDefs.F_DELETEEMPTYTRACKS, StrNoneChk(self.jet_file.config.delete_empty_tracks))
        
        result = dlg.ShowModal()
        if result == wx.ID_OK:
            self.jet_file.config.filename = dlg.je.ctrls[JetDefs.F_JETFILENAME].GetValue()
            self.jet_file.config.copyright = dlg.je.ctrls[JetDefs.F_COPYRIGHT].GetValue()
            self.jet_file.config.chase_controllers = dlg.je.ctrls[JetDefs.F_CHASECONTROLLERS].GetValue()
            self.jet_file.config.delete_empty_tracks = dlg.je.ctrls[JetDefs.F_DELETEEMPTYTRACKS].GetValue()
        dlg.Destroy()
       
    def OnPreferences(self, event):
        """ Opens a dialog box to capture preferences and saves them to a configuration file """
        dlg = JetPreferences()
        dlg.SetValue(JetDefs.F_COPYRIGHT, IniGetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_PREF_SECTION, JetDefs.F_COPYRIGHT))
        dlg.SetValue(JetDefs.F_CHASECONTROLLERS, IniGetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_PREF_SECTION, JetDefs.F_CHASECONTROLLERS, 'bool', True))
        dlg.SetValue(JetDefs.F_DELETEEMPTYTRACKS, IniGetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_PREF_SECTION, JetDefs.F_DELETEEMPTYTRACKS, 'bool', False))
        result = dlg.ShowModal()
        if result == wx.ID_OK:
            IniSetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_PREF_SECTION, JetDefs.F_COPYRIGHT, dlg.GetValue(JetDefs.F_COPYRIGHT))
            IniSetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_PREF_SECTION, JetDefs.F_CHASECONTROLLERS, dlg.GetValue(JetDefs.F_CHASECONTROLLERS))
            IniSetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_PREF_SECTION, JetDefs.F_DELETEEMPTYTRACKS, dlg.GetValue(JetDefs.F_DELETEEMPTYTRACKS))
        dlg.Destroy()

    def LoadDefaultProperties(self):
        """ Loads default properties from the a configuration file """
        self.jet_file.config.copyright = IniGetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_PREF_SECTION, JetDefs.F_COPYRIGHT)
        self.jet_file.config.chase_controllers = IniGetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_PREF_SECTION, JetDefs.F_CHASECONTROLLERS, 'bool', True)
        self.jet_file.config.delete_empty_tracks = IniGetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_PREF_SECTION, JetDefs.F_DELETEEMPTYTRACKS, 'bool', False)               
                
    def OnClose(self, event):
        """ Called upon closing the JetCreator main window """
        if self.isDirty():
            ret = YesNoCancel(JetDefs.MAIN_JETCREATOR, JetDefs.MAIN_SAVEBEFOREEXIT, True)
            if ret == wx.ID_YES:
                self.OnJetSave(None)
            if ret == wx.ID_CANCEL:
                return
            
        if self.jet is not None:
            SafeJetShutdown(self.playerLock, self.jet)                
        self.Destroy()
        
    def OnPlay(self, event):
        """ Plays the currently queued segments """
        if self.btnPlay.GetLabel() == JetDefs.BUT_PLAY:
            if not ValidateConfig(self.jet_file):
                return
            
            #make sure something is queued
            iQueued = False
            num = self.segList.GetItemCount()
            for seg_num in range(num):
                if self.segList.IsChecked(seg_num):
                    iQueued = True
            if not iQueued:
                InfoMsg(JetDefs.MAIN_PLAYSEG, JetDefs.MAIN_PLAYSEGMSG)
                return
        
            for segment in self.jet_file.segments:
                if FileExists(segment.dlsfile):
                    if not segment.dlsfile in self.jet_file.libraries:
                        self.jet_file.libraries.append(segment.dlsfile)
                    
            self.eventList.DeleteAllItems()
            num = self.segList.GetItemCount()
            for seg_num in range(num):
                if seg_num == 0: self.log.Clear()
                if self.segList.IsChecked(seg_num):
                    segment = self.jet_file.GetSegment(getColumnText(self.segList, seg_num, 0))
                    if segment != None:
                        #so we can determine which segment is playing, make these the same
                        userID = seg_num
                        if FileExists(segment.dlsfile):
                            dls_num = FindDlsNum(self.jet_file.libraries, segment.dlsfile)
                        else:
                            dls_num = -1
                        self.queueSegs.append(QueueSeg(segment.segname, userID, seg_num, dls_num, segment.repeat, segment.transpose, segment.mute_flags))
                    
            if len(self.queueSegs) == 0:
                return
            
            self.btnPlay.SetLabel(JetDefs.BUT_STOP)
            
            thread.start_new_thread(self.PlaySegs, ())
        else:
            with self.playerLock:
                self.jet.Clear_Queue()
            self.SetKeepPlayingFlag(False)
    
    def PlaySegs(self):
        """ Thread writes a temporary copy of the jet file, and calls the library to play it """
        if len(self.queueSegs) == 0:
            return
        
        jet_file = CreateTempJetFile(self.jet_file)
        
        if not ValidateConfig(jet_file):
            CleanupTempJetFile(jet_file)
            return
             
        self.jet = JET()
        self.jet.eas.StartWave()
        self.jet.OpenFile(jet_file.config.filename)

        lastID = -1

        # queue first segment and start playback
        Queue(self.jet, self.queueSegs[0])
        index = 1
        self.jet.Play()
        self.paused = False
    
        # continue playing until all segments are done
        self.SetKeepPlayingFlag(True)
        while self.GetKeepPlayingFlag():
            self.jet.Render()
            status = self.jet.Status()
            
            if status.currentUserID <> lastID and status.currentUserID <> -1:
                wx.PostEvent(self, JetStatusEvent(JetDefs.PST_PLAY, status.currentUserID))
                lastID = status.currentUserID

            # if no more segments - we're done
            if status.numQueuedSegments == 0:
                break
    
            self.jet.GetAppEvent()
    
            # if less than 2 segs queued - queue another one            
            if (index < len(self.queueSegs)) and (status.numQueuedSegments < 2):
                Queue(self.jet, self.queueSegs[index])
                index += 1
                
            wx.PostEvent(self, JetStatusEvent(JetDefs.PST_UPD_LOCATION, status.location))
        
        
        SafeJetShutdown(self.playerLock, self.jet)   
        
        self.queueSegs = []
        
        CleanupTempJetFile(jet_file)
        
        wx.PostEvent(self, JetStatusEvent(JetDefs.PST_DONE, None))
                
    def OnJetStatusUpdate(self, evt):
        """ These are screen updates called for from within the thread.  Communication 
            is via a postevent call.
        """
        if evt.mode == JetDefs.PST_PLAY:
            segName = getColumnText(self.segList, evt.data, 0)
            self.LoadEventsForSeg(segName)
            self.log.SetValue(segName)
            ClearRowSelections(self.segList)
            SetRowSelection(self.segList, evt.data, True)
        elif evt.mode == JetDefs.PST_UPD_LOCATION:
            self.graph.UpdateLocation(evt.data)
        elif evt.mode == 3:
            self.graph.UpdateLocation(0)
            ClearRowSelections(self.segList)
            self.eventList.DeleteAllItems()
            self.SetKeepPlayingFlag(False)
            self.btnPlay.SetLabel(JetDefs.BUT_PLAY)
            self.btnPause.SetLabel(JetDefs.BUT_PAUSE)            
            
    def OnPause(self, evt):
        """ Pauses the playback """
        if self.jet is None:
            return
        if not self.paused:
            self.jet.Pause()
            self.paused = True
            self.btnPause.SetLabel(JetDefs.BUT_RESUME)
        else:
            self.jet.Play()
            self.paused = False
            self.btnPause.SetLabel(JetDefs.BUT_PAUSE)
           
    def isDirty(self):
        if len(self.UndoStack) == 0 and len(self.RedoStack) == 0:
            return False
        else:
            return True
        
    def OnSetGraphOptions(self, evt):
        """ Sets graph options """
        IniSetValue(self.currentJetConfigFile, JetDefs.F_GRAPHLABELS, JetDefs.F_GRAPHLABELS, self.chkGraphLabels.GetValue())
        IniSetValue(self.currentJetConfigFile, JetDefs.F_GRAPHCLIPS, JetDefs.F_GRAPHCLIPS, self.chkGraphClips.GetValue())
        IniSetValue(self.currentJetConfigFile, JetDefs.F_GRAPHAPPEVTS, JetDefs.F_GRAPHAPPEVTS, self.chkGraphAppEvts.GetValue())
        self.graph.LoadSegment(self.jet_file.GetSegment(self.currentSegmentName), showLabels=IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHLABELS, JetDefs.F_GRAPHLABELS, 'bool', 'True'), showClips=IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHCLIPS, JetDefs.F_GRAPHCLIPS, 'bool', 'True'), showAppEvts=IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHAPPEVTS, JetDefs.F_GRAPHAPPEVTS, 'bool', 'True'))

    def OnEventSortOrderChanged(self):
        """ Called when sort order has changed """
        IniSetValue(self.currentJetConfigFile, JetDefs.INI_EVENTSORT, JetDefs.INI_EVENTSORT_0, self.eventList.GetSortState()[0])
        IniSetValue(self.currentJetConfigFile, JetDefs.INI_EVENTSORT, JetDefs.INI_EVENTSORT_1, self.eventList.GetSortState()[1])
        self.LoadEventsForSeg(self.currentSegmentName)
        self.graph.LoadSegment(self.jet_file.GetSegment(self.currentSegmentName), showLabels=IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHLABELS, JetDefs.F_GRAPHLABELS, 'bool', 'True'), showClips=IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHCLIPS, JetDefs.F_GRAPHCLIPS, 'bool', 'True'), showAppEvts=IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHAPPEVTS, JetDefs.F_GRAPHAPPEVTS, 'bool', 'True'))

    def EventSortCmp(self, a, b):
        """ Sorts based on selected sort order """
        if self.eventlistSort[0] == 0 or self.eventlistSort[0] == 1:
            if self.eventlistSort[1] == 1:
                return cmp(a[0].upper(), b[0].upper())
            else:
                return cmp(b[0].upper(), a[0].upper())                
        elif self.eventlistSort[0] == 2 or self.eventlistSort[0] == 3: 
            if self.eventlistSort[1] == 1:
                return cmp(MbtVal(a[0]), MbtVal(b[0]))
            else:
                return cmp(MbtVal(b[0]), MbtVal(a[0]))   
        else:
            return cmp(a[0], b[0])             

    def EventSort2(self, seq, attr):
        """ Does Sorting """
        intermed = map(None, map(getattr, seq, (attr,)*len(seq)), xrange(len(seq)), seq)
        intermed.sort(self.EventSortCmp)
        return map(operator.getitem, intermed, (-1,) * len(intermed))

    def EventSort(self, lst, attr):
        """ Does Sorting """
        lst[:] = self.EventSort2(lst, attr)
        
    def OnSegSortOrderChanged(self):
        """ Called when sort order has changed """
        IniSetValue(self.currentJetConfigFile, JetDefs.INI_SEGSORT, JetDefs.INI_SEGSORT_0, self.segList.GetSortState()[0])
        IniSetValue(self.currentJetConfigFile, JetDefs.INI_SEGSORT, JetDefs.INI_SEGSORT_1, self.segList.GetSortState()[1])
        self.LoadEventsForSeg(self.currentSegmentName)
        self.graph.LoadSegment(self.jet_file.GetSegment(self.currentSegmentName), showLabels=IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHLABELS, JetDefs.F_GRAPHLABELS, 'bool', 'True'), showClips=IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHCLIPS, JetDefs.F_GRAPHCLIPS, 'bool', 'True'), showAppEvts=IniGetValue(self.currentJetConfigFile, JetDefs.F_GRAPHAPPEVTS, JetDefs.F_GRAPHAPPEVTS, 'bool', 'True'))

    def SegSortCmp(self, a, b):
        """ Sorts based on selected sort order """
        if self.seglistSort[0] == 0:
            if self.seglistSort[1] == 1:
                return cmp(a[0].upper(), b[0].upper())
            else:
                return cmp(b[0].upper(), a[0].upper())   
        elif self.seglistSort[0] == 1 or self.seglistSort[0] == 2:   
            if self.seglistSort[1] == 1:
                return cmp(FileJustName(a[0]).upper(), FileJustName(b[0]).upper())
            else:
                return cmp(FileJustName(b[0]).upper(), FileJustName(a[0]).upper())   
        elif self.seglistSort[0] == 3 or self.seglistSort[0] == 4: 
            if self.seglistSort[1] == 1:
                return cmp(MbtVal(a[0]), MbtVal(b[0]))
            else:
                return cmp(MbtVal(b[0]), MbtVal(a[0]))   
        else:
            return cmp(a[0], b[0])             

    def SegSort2(self, seq, attr):
        """ Does Sorting """
        intermed = map(None, map(getattr, seq, (attr,)*len(seq)), xrange(len(seq)), seq)
        intermed.sort(self.SegSortCmp)
        return map(operator.getitem, intermed, (-1,) * len(intermed))

    def SegSort(self, lst, attr):
        """ Does Sorting """
        lst[:] = self.SegSort2(lst, attr)

if __name__ == '__main__':
    """ Sets the logging level, then calls the open dialog box before initializing main window"""
    
    logLevel = IniGetValue(JetDefs.JETCREATOR_INI, JetDefs.INI_LOGGING, JetDefs.INI_LOGGING)
    if logLevel == 'DEBUG':
        logging.basicConfig(level=logging.DEBUG, format='%(funcName)s[%(lineno)d]: %(message)s')
    elif logLevel == 'INFO':
        logging.basicConfig(level=logging.INFO, format='%(funcName)s[%(lineno)d]: %(message)s')
    elif logLevel == 'ERROR':
        logging.basicConfig(level=logging.ERROR, format='%(funcName)s[%(lineno)d]: %(message)s')
    elif logLevel == 'WARNING':
        logging.basicConfig(level=logging.WARNING, format='%(funcName)s[%(lineno)d]: %(message)s')
    else:
        install_release_loggers()
        
    app = wx.App(None)
    
    if not os.path.isdir(JetDefs.TEMP_JET_DIR):
        os.mkdir(JetDefs.TEMP_JET_DIR)
        
    openFile = ""
    dlg = JetOpen()
    result = dlg.ShowModal()
    if result == wx.ID_CANCEL:
        dlg.Destroy()
    elif result == JetDefs.ID_JET_IMPORT:
        dlg.Destroy()
    
        au = JetCreator(None, -1, "", importFlag=True)
        app.MainLoop()
    else:
        openFile = dlg.fileName
        dlg.Destroy()
    
        au = JetCreator(None, -1, openFile)
        app.MainLoop()




#----------------------------------------------------------------------
# This file was generated by C:\Python25\Lib\site-packages\wx-2.8-msw-unicode\wx\tools\img2py.py
#
from wx import ImageFromStream, BitmapFromImage, EmptyIcon
import cStringIO, zlib


def getData():
    return zlib.decompress(
'x\xda\x01\xdd\x01"\xfe\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00 \x00\
\x00\x00 \x08\x06\x00\x00\x00szz\xf4\x00\x00\x00\x04sBIT\x08\x08\x08\x08|\
\x08d\x88\x00\x00\x01\x94IDATX\x85\xed\x97\xbdN\xc30\x14\x85OJ\x06gK6\x18;\
\xa6\x1bk^\xa1\x8fS\xb5\x08\x18\x10\x03c\x07\x84X\x902\xb6o\x90\x951\xdd\xca\
\xd6\x8c\x1e\x93\xcd\x06\x06\xdf\xad\x0c\xb4\x15HIs\x9dDD\x958[,\xfb\x9e/\
\xbe\xd7\x7f\x8e38C\x9fr\x01\xe0\xe9\xe5qKZw\x1cZ`:\xbbrX\x00\xa45&\x93\xdbN\
\xed\xe7\xf3\x07V?\xf7\xe7\x87\xeeh\x16|\xdfg\xf7\x1dt\xe2\xd8B\xff\x00nYc\
\x96e\x8d\x82\r\x87C\x08!\xda\x03\x84a\xd8\x08\xa0\x89N#\x05i\x9aB)\xb2\x0e~\
s}\xbf\x05\x80 \x00\xa6\xb3\xbb\xd2M\x89\x95\x82$y\x85\x94\xd6\xfe,\x9dF\n\
\xfe\x1c\xa0,\x05{\x8dF\xe7\x10\xc2\xab\rLd\xb0\xd9\x14\xcd\x00\x8e\xc9\x18\
\x821\x9c\x9e\xbc\xa2\xb5N\x81\xd6\xc4\nN\xccUS\x99\x02\xa2\xf2\x00ax\xc1\
\xda\xedV+\t\x98z\x88R\x80*s\x00H\xd3n\xd7\xe3/\x80c\xc6{\x8d\xc7\x97\xa8\
\x9b\x80\xf5ZBJ\xde\xdd\xc2\xba\x08\x93\xe4\xcdv\x08\x0f\x80\xf3\xf7\x00\x10\
E\xbc\x13\x8f\x88X\xe9\xb2\x9e\x81,\xcbm\x87t\x0b\xd0\xe4Pb\x01\x10\x91\xf5e\
\xa2L\x9e\x07\xf8>v\x85*vm\xd5\xd0;\x00\x818~nm\x0e|\x1bGQx\xb8\x19\x17E\x81\
<\xaf\xae\x05\x17@\xed\x03b\x7f\xae\xf3!\xc4\xc1\\)\x05)\xab\xeb\xc6\xba\x06\
8\xd2Z\xc3\x18\x03M\x1aRJ,\x16\x8bv\x00A\xc07\xf7< \xcfs(RX\xc6K(\xa5\xf0\
\xfe\xf1Y9\xc3N\xdf\x8f\xd3\xdeoD\xbd\x03|\x01\xd0i\x97\xf5j]\xc4\xe0\x00\
\x00\x00\x00IEND\xaeB`\x82q\x05\xc6\xef' )

def getBitmap():
    return BitmapFromImage(getImage())

def getImage():
    stream = cStringIO.StringIO(getData())
    return ImageFromStream(stream)

def getIcon():
    icon = EmptyIcon()
    icon.CopyFromBitmap(getBitmap())
    return icon




#----------------------------------------------------------------------
# This file was generated by C:\Python25\Lib\site-packages\wx-2.8-msw-unicode\wx\tools\img2py.py
#
from wx import ImageFromStream, BitmapFromImage, EmptyIcon
import cStringIO, zlib


def getData():
    return zlib.decompress(
'x\xda\x01\xd4\x03+\xfc\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00 \x00\
\x00\x00 \x08\x06\x00\x00\x00szz\xf4\x00\x00\x00\x04sBIT\x08\x08\x08\x08|\
\x08d\x88\x00\x00\x03\x8bIDATX\x85\xc5\x97klSe\x18\xc7\x7f\xef9\xed\xda\xb5\
\xdd\xa8\xeb\xec\x9cd\xc9\xe6\xb0&\x18#\xc6(\xd1\xb0(\n\x0eG\xb8\x05GLX\xe2\
\x17\xa3\x04\xd4hP\x13t&\x98\r\xbf\x90h\xe2\x06Y\xf8`d\x18!r\t\xa0f\xb0\xa0s\
\xea\xbc\x8d\xc5!c\x85Ywi\xe9\xd8\xad[\xaf\xf3\xf4z\xfc\xe2t\x10l\xc7\xba\
\xd6\xff\xc7s\x9e\xff\xf3\xfe\xce\xfb\xbe\xcf\x93\xe7\x08!\xc9\xfc\x9f\xd2\
\x00$\xe21u\xbe\t\xa2\xfc\x8e\x8a\x96\x1cJ\xe7\xe5\x97\xe6\xbb0\x80\xcb\xe9b\
[\xfdS\xf4xN\xcf;GZ\x00\xa1\xa0\x82\xe3\xd28\xae\xc0\xb7\xc0\xfc61-\x00I\x96\
\xc83\xdc\xc6h\xf0\x02\xed\x83u\xd9\x07\x00PU\x15\x93\xc1\x8c\xdd\x7f\x90\
\xf6\xc1\xfa\xec\x03\xe8\x8d2mGGq\xfc,\xe8\x8f6\xf3\xd5\xc0.Tu\xee\xc7\x916\
\x80\x90\x12X,\x054\xbd:\xc2\xe9F\x0f\xfd\xcag\xb4\xbbj\xb3\x070\x1d\x88\xf3\
P\x95\x91\xc3g\xdf\xe3\xf2\x17e4l\xbf\x8a}\xeaSz=G\xb2\x03 !\x11\x0c\x06\x88\
\x96\xb4\xd0\xda\xd5\x88nb\x05Mo8\xe9\xf1\x1f@\x89y\xb3\x00 \x0b\xc6\x06TF\
\xc6\x878\x1f|\x87\x83gw\xe2\xe9\xb6\xd1r\xbc\x93\xf1\xf8\x8f\x99\x070\xe4\
\xcbt\x9d\xf3\xb1\xa7\xe6\n=\xdd\x7fp5\xaf\x81\x17k+\xe8\xfa2\x8e\'|)\xf3\
\x00\x01\x9fB\xf5\xf6{yv\xfd\xcb\xec\xd92\xc4\xe7\x87z(\xad\xec\xa6\xd0\x9a\
\xcf\xc4X\xea#\xd0\xa4\x0b@B\x837<\xc8\x0bu\x1bX\xf9h\x0b\xeb7\xadE\xda/\xb1\
\xe95\x0b"\x91\x93\xd2\xbe \x8d\xc8\xa03\xf1M\xdf\x07\xdc\xb7\xd6\xc3\xf1c\'\
i|\xc5I(\xe4\xe7N\xeb\x92\xcc\x03\x18\xf2\xb4\x1cmt`\xef\x08\xd36\xb4\x9b\
\x077O\xb2\xb5\xfay\x8e5\xb8\x91\xf5J\xa6\x01T\x0cfXv\xd7j\x9a\xdf\xf2\xd1\
\xbc\xdb\xcd\xd7}{Y\xb9-\x84\x7f8\x07\xf7x_\xa6\x01\xc0\xe7\xf3\xf2\xfa\x8e\
\x9dtv\x9e\'2\xf0\x00o>}\x11\xaf\xda\xcd\x86\x97,(\x81\xd4\xfe\xf4[\xb1\x10L\
(\x1e\xca\xcb\xcb\xf9\xbe\xe3;\xaa+w\xf0\xf6F;\xd6R\r\x8b\x8b\xeeN\xe9O\xbf\
\n\xfe\x86\x98\xd1\xbe\xfd\x8d\xa0\n>\xaam\xa2\xe2d?**\x02\xf1\x9f\xde\xb4w\
\xe0f\xfap\xdf\xfb\x14\xe9\x96\xd1\xf5k\'q\x82Ic3\x02 KZ\xaa\xd6l\xe4\xda`\
\x08A$\xfb\x00\x006\xdb\x124\x1a\x19\t}\xe6\x00\xb4\xc2D4|\xf3\xe1cl\xca\x89\
>W\x87\xc0\x984\xc7-_\xc2\xc9?\x1d8&\xcf\xe0\x8f\xb8\xf0\xfb\xe2\x94\xd8\x0c\
\xc4c\x80\xee\xfa8\x87\xfb\'V\xad{2e\xbe[\x02\xe8p\xee\xe5\xb7\xd1C\x84\xe3\
\x01$\xa1A\x96%\xee\x7f\xdcDTQ\x99\xfd\xa1\xc3\xc3nb\xdaQ*\x96>\xb70\x00*\tZ\
\xaf\xec\xa2w\xea0\xb9\x9a\x02\x0cZ\xcb?\xef\x94\x88\x97\x1bG\xf2~\xd7E\xb6\
\xae\xa9\'W*^\x18\x80\x8fO\xd43b\xfe\x84Bs1\xb3\xe7M\x95\x04\x92\xd0\xb0H_z]\
\xfc\x8a\xe5\x95\x90\xa4\xf6g+\xe5%\xbc\xe6\x1e\xe5\xc8\x89\x03\xdcQl%\x91\
\xf8\xf7\xb9\x10\x02%\xe6\xc1fY\xc7\xed\xc6\xa57\xb8\xe6\xb6\xf8\x9c\x00"\
\xb1i.\xfc0\xc2\xb47\x81\xb9PK\xaeIFgT\x99\xf4\r\xb38g5O\x94\xd5%\xedt\xa9$\
\x84$\'\xfd9U\xd58\xcfTo\xa6\xed\x97S\xac\xaa\xc9gQA\x0e\t\xc5\xc4\xf2{\xb6P\
S\xf5.\x06c\xf2:O\x1b`F\xad\xe7\xce`\xef\xbdLiY\t\x8f<\xfc\x18\xd6\xa2\xc2\
\xb4\x16\x9e\xd1_\xa4\x9cO6P\x7f\x0e\x9e\x00\x00\x00\x00IEND\xaeB`\x82\x80\
\xab\xe3<' )

def getBitmap():
    return BitmapFromImage(getImage())

def getImage():
    stream = cStringIO.StringIO(getData())
    return ImageFromStream(stream)

def getIcon():
    icon = EmptyIcon()
    icon.CopyFromBitmap(getBitmap())
    return icon




"""
 File:  
 JetStatusEvent.py
 
 Contents and purpose:
 Creates an event for postevent callbacks
 
 Copyright (c) 2008 Android Open Source Project
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

import wx

EVT_JET_STATUS_ID = wx.NewId()

def EVT_JET_STATUS(win, func):
    win.Connect(-1, -1, EVT_JET_STATUS_ID, func)

class JetStatusEvent(wx.PyEvent):
    """Used for posting events out of play thread back to UI"""
    def __init__(self, mode, data):
        wx.PyEvent.__init__(self)
        self.SetEventType(EVT_JET_STATUS_ID)
        self.mode = mode
        self.data = data


